"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import inspect
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union, Literal, Iterable
from pathlib import Path
from datetime import datetime, timezone, timedelta

# * Third Party Imports --------------------------------------------------------------------------------->
from tzlocal import get_localzone

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.enums import LoggingLevel, LoggingSectionAlignment
from gidapptools.general_helper.string_helper import StringCase, StringCaseConverter

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_logger.records import LOG_RECORD_TYPES

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def get_all_func_names(file: Path, recursive: bool = True):
    import ast

    def _get_names(source_file: Path) -> list[str]:
        root = ast.parse(source_file.read_text(encoding='utf-8', errors='ignore'))
        return sorted({node.name for node in ast.walk(root) if isinstance(node, ast.FunctionDef)})

    def _process_item(_item: Path):
        if _item.is_file() and _item.suffix == '.py':

            yield _get_names(_item)

        elif _item.is_dir():
            for _sub_item in _item.iterdir():
                yield from _process_item(_sub_item)
    if recursive is False:
        names = _get_names(file)
    else:
        names = []
        for _names in _process_item(file.parent):
            names += _names
    return sorted(set(names), key=len)


def get_all_module_names(file: Path) -> list[str]:

    def _get_module_name(_item: Path) -> str:
        module_name = _item.stem
        return module_name

    def _process_item(_item: Path):
        if _item.is_file() and _item.suffix == '.py':
            yield _get_module_name(_item)
        elif _item.is_dir():
            for _sub_item in _item.iterdir():
                yield from _process_item(_sub_item)

    if file.is_file():
        file = file.parent
    return sorted({i for i in _process_item(file)}, key=len)


class AbstractLoggingStyleSection:
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT
    default_width: int = 0
    default_text: str = str(None)

    __slots__ = ("position",
                 "width",
                 "alignment")

    def __init__(self,
                 position: int = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        self.position = position
        self.width = width or self.default_width

        if alignment is None:
            self.alignment = self.default_alignment
        elif isinstance(alignment, str):
            self.alignment = LoggingSectionAlignment(alignment)
        else:
            self.alignment = alignment

    @abstractmethod
    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        ...

    def align(self, text: str) -> str:
        width = self.width() if callable(self.width) else self.width
        if len(text) > width:
            text = '...' + text[-(width - 3):]
        return self.alignment.align(text, width)

    def format(self, record: "LOG_RECORD_TYPES") -> str:
        try:
            text = str(self.get_formated_value(record=record))
        except AttributeError:
            text = self.default_text
        return self.align(text)

    @classmethod
    def ___section_name___(cls) -> str:
        return StringCaseConverter.convert_to(cls.__name__.removesuffix("Section"), StringCase.SNAKE)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(position={self.position!r})"


class TimeSection(AbstractLoggingStyleSection):

    local_timezone: timezone = get_localzone()
    default_time_format = '%Y-%m-%d %H:%M:%S'
    default_msec_format = '.{msec:03.0f}'
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT

    __slots__ = ("time_zone",
                 "time_format",
                 "msec_format")

    def __init__(self,
                 position: int = None,
                 time_zone: timezone = None,
                 time_format: str = None,
                 msec_format: str = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, alignment=alignment)

        self.time_zone = time_zone
        self.time_format = time_format or self.default_time_format
        self.msec_format = msec_format or self.default_msec_format

        self.width = width or (len(self.time_format) + len(self.msec_format.format(msec=0)) + self._get_time_zone_width() + 2)

    def _get_time_zone_width(self) -> int:
        time_zone_width = 1
        if self.time_zone is not None:
            time_zone_width += len(self.time_zone.tzname(datetime.now()))
        else:
            time_zone_width += len(get_localzone().tzname(datetime.now()))

        return time_zone_width

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        time_value = datetime.fromtimestamp(record.created, tz=self.local_timezone)
        msec_value = record.msecs
        msec_string = self.msec_format.format(msec=msec_value).strip()
        if msec_string == ".1000":
            msec_string = ".000"
            time_value = time_value + timedelta(seconds=1)
        if self.time_zone is not None:
            time_value = time_value.replace(tzinfo=self.time_zone)
        if self.time_format.casefold() == "isoformat":
            return time_value.isoformat(timespec='seconds') + self.msec_format.format(msec=msec_value)
        return (time_value.strftime(self.time_format) + msec_string + time_value.strftime(" %Z")).strip()


class LevelSection(AbstractLoggingStyleSection):
    default_case: StringCase = StringCase.UPPER
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT_EXTRA_PADDED_ONCE
    default_width: int = max(len(i) for i in LoggingLevel._member_names_) + 2
    __slots__ = ("case",
                 "_formated_value_cache")

    def __init__(self,
                 position: int = None,
                 case: StringCase = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, alignment=alignment, width=width)
        if case is None:
            self.case = self.default_case
        elif isinstance(case, str):
            self.case = StringCase(case)
        else:
            self.case = case

        self._formated_value_cache: dict[str, str] = {}

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        try:
            is_qt = record.extras["is_qt"]

        except (AttributeError, KeyError):
            is_qt = False

        raw_level_name = record.levelname

        if is_qt is True:
            raw_level_name = "Qt_" + raw_level_name

        try:
            return self._formated_value_cache[raw_level_name]
        except KeyError:
            formated_level_name = StringCaseConverter.convert_to(raw_level_name, self.case)
            self._formated_value_cache[raw_level_name] = formated_level_name

            return formated_level_name


class ThreadSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.CENTER
    default_width: int = 20
    __slots__ = tuple()

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        return record.threadName


class LineNumberSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.CENTER
    default_width: int = 5
    __slots__ = tuple()

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        return record.lineno


class PathSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT
    __slots__ = ("with_extension",
                 "_width_cache")

    def __init__(self,
                 position: int = None,
                 with_extension: bool = True,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, width=self.width_from_env, alignment=alignment)
        self.with_extension = with_extension
        self._width_cache: int = width or None

    def width_from_env(self) -> int:
        if self._width_cache is None:
            self._width_cache = int(os.getenv("MAX_MODULE_NAME_LEN", "20"))
        return self._width_cache

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        if record.extras.get("file_path", None) is not None:
            path = Path(record.extras["file_path"]).resolve()
        else:
            path = Path(record.pathname).resolve()
        if os.getenv("_MAIN_DIR", None) is not None:
            path = path.relative_to(Path(os.getenv("_MAIN_DIR")).resolve())
        if self.with_extension is False:
            path = path.with_suffix('')

        return './' + path.as_posix()


class LoggerNameSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT
    default_width: int = 20
    __slots__ = tuple()

    def __init__(self,
                 position: int = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position=position, width=width, alignment=alignment)

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        if record.extras.get("module", None) is not None:
            return record.extras["module"]
        return record.name


class FunctionNameSection(AbstractLoggingStyleSection):
    default_alignment: LoggingSectionAlignment = LoggingSectionAlignment.LEFT
    default_width: int = 20
    default_text: str = "-"
    __slots__ = tuple()

    def __init__(self, width: int = None) -> None:
        super().__init__(width=width)

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        if record.funcName == "<module>":
            return self.default_text
        if record.extras.get("function_name", None) is not None:
            return record.extras["function_name"]
        return record.funcName


class ExtrasSection(AbstractLoggingStyleSection):
    __slots__ = tuple()

    def __init__(self) -> None:
        super().__init__(width=100)

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:
        try:
            extras = record.extras
        except AttributeError:
            return ""

        return ' + '.join(f"{k}={v!r}" for k, v in extras.items())


class PackageNameSection(AbstractLoggingStyleSection):
    __slots__ = tuple()

    def __init__(self,
                 position: int = None,
                 width: int = None,
                 alignment: Union[LoggingSectionAlignment, str] = None) -> None:
        super().__init__(position, width or 12, alignment)

    def get_formated_value(self, record: "LOG_RECORD_TYPES") -> str:

        name_parts = record.name.split('.')

        package_name = name_parts[0]

        try:
            if record.extras.get("is_qt", False) is True:
                package_name = "Qt"
        except AttributeError:
            pass
        return package_name


class AbstractSectionLoggingStyle(ABC):
    default_separator: str = " | "
    default_message_start_indicator: str = ' ||--> '
    __slots__ = ("sections",
                 "sorted_sections",
                 "separator",
                 "message_start_indicator",
                 "message_section")

    def __init__(self, sections: Iterable[Union[str, AbstractLoggingStyleSection]], separator: str = None, message_start_indicator: str = None, message_section: AbstractLoggingStyleSection = None):
        self.sections = self._handle_sections(sections)
        self.sorted_sections: tuple[AbstractLoggingStyleSection] = self.sort_sections()
        self.separator = separator or self.default_separator
        self.message_start_indicator = message_start_indicator or self.default_message_start_indicator
        self.message_section = message_section

    def _handle_sections(self, sections: Iterable[Union[str, AbstractLoggingStyleSection]]) -> tuple[AbstractLoggingStyleSection]:
        _out = []
        for section in sections:
            if isinstance(section, str):
                section = self.all_section_classes[section]()
            if inspect.isclass(section):
                section = section()
            _out.append(section)
        return tuple(_out)

    def sort_sections(self) -> tuple[AbstractLoggingStyleSection]:
        temp_sorted = [i[0] for i in sorted([(v, idx) for idx, v in enumerate(self.sections)], key=lambda x: (x[0].position, x[1]))]
        for idx, sect in enumerate(temp_sorted):
            if sect.position is None:
                sect.position = idx
        return tuple(temp_sorted)

    def usesTime(self) -> bool:
        return False

    @ abstractmethod
    def validate(self) -> None:
        ...

    @ abstractmethod
    def _format(self, record: "LOG_RECORD_TYPES") -> str:
        ...

    def format(self, record: "LOG_RECORD_TYPES"):
        return self._format(record)

    @classmethod
    def all_section_classes(cls) -> dict[str, type[AbstractLoggingStyleSection]]:
        _out = {}
        for klass in AbstractLoggingStyleSection.__subclasses__():
            _out[klass.___section_name___] = klass
        return _out


class GidSectionLoggingStyle(AbstractSectionLoggingStyle):
    __slots__ = tuple()

    def validate(self) -> None:
        if any(isinstance(i, AbstractLoggingStyleSection) is False for i in self.sections):
            raise ValueError("invalid format")

    def _format(self, record: "LOG_RECORD_TYPES") -> str:

        message_part = record.getMessage() if self.message_section is None else self.message_section.format(record)

        return self.separator.join(section.format(record) for section in self.sorted_sections) + self.message_start_indicator + message_part


class GidLoggingFormatter(logging.Formatter):
    default_fmt = (TimeSection(), LineNumberSection(), LevelSection(), ThreadSection(), PackageNameSection(), LoggerNameSection(), FunctionNameSection())

    def __init__(self,
                 fmt: Union[str, Iterable[Union[str, AbstractLoggingStyleSection]]] = None,
                 style: Literal["section", "$", "{", "$"] = "section",
                 validate: bool = False,
                 **kwargs) -> None:
        if style == "section":
            if fmt is None:
                fmt = self.default_fmt
            self._style = GidSectionLoggingStyle(sections=fmt, separator=kwargs.pop("separator", None), message_start_indicator=kwargs.pop("message_start_indicator", None), message_section=kwargs.pop("message_section", None))
        else:
            self._style = logging._STYLES[style][0](fmt)
            if validate:
                self._style.validate()

            self._fmt = self._style._fmt
            self.datefmt = kwargs.pop("datefmt", None)

    def set_time(self, record: "LOG_RECORD_TYPES") -> "LOG_RECORD_TYPES":
        return record

    def format_message(self, record: "LOG_RECORD_TYPES") -> str:
        return self.formatMessage(record=record)

    def format_stack(self, stack_info):
        return self.formatStack(stack_info=stack_info)

    def format_exception(self, record: "LOG_RECORD_TYPES") -> "LOG_RECORD_TYPES":
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(ei=record.exc_info)
        return record

    def modify_meta_record(self, record: "LOG_RECORD_TYPES") -> "LOG_RECORD_TYPES":

        if getattr(record, "is_meta", False) is False:
            return record

        record.name = "__META__"
        record.module = "-"
        record.funcName = "-"
        record.lineno = "-"
        extras = getattr(record, "extras", {})
        extras["module"] = "-"
        record.extras = extras
        return record

    def format(self, record: "LOG_RECORD_TYPES") -> str:
        record = self.set_time(record=record)
        record = self.modify_meta_record(record)
        text = self.format_message(record=record)
        record = self.format_exception(record=record)
        if record.exc_text:
            if text[-1:] != "\n":
                text = text + "\n"
            text = text + record.exc_text
        if record.stack_info:
            if text[-1:] != "\n":
                text = text + "\n"
            text = text + self.format_stack(record.stack_info)
        return text

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from typing import Union, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtGui import QIcon, QPixmap
from PySide6.QtCore import Qt, QSize
from PySide6.QtWidgets import QLabel, QBoxLayout, QHBoxLayout, QPushButton, QVBoxLayout, QApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ImageTextButton(QPushButton):

    def __init__(self, icon: QPixmap = None, text: str = None, direction: Union[type[Qt.Vertical], type[Qt.Horizontal]] = Qt.Vertical, parent=None):
        super().__init__(parent=parent)
        layout = QVBoxLayout() if direction == Qt.Vertical else QHBoxLayout()
        self.setLayout(layout)
        self.text_label = QLabel()
        self.text_label.setAlignment(Qt.AlignCenter)
        self.set_text(text)
        self.layout.addWidget(self.text_label)

        self.icon_label = QLabel()
        self.icon_label.setAlignment(Qt.AlignCenter)
        self.set_icon(icon)
        self.layout.addWidget(self.icon_label)

    @property
    def direction(self):
        if isinstance(self.layout, QVBoxLayout):
            return Qt.Vertical
        return Qt.Horizontal

    @property
    def label_text(self) -> Optional[str]:
        return self.text_label.text()

    @property
    def icon(self) -> Optional[QIcon]:
        return QIcon(self.icon_label.pixmap())

    def _get_correct_image_size(self) -> QSize:
        if not self.label_text:

            return self.sizeHint()

        fm = self.text_label.fontMetrics()
        widths = []
        heights = []
        for line in self.label_text.splitlines():
            br = fm.boundingRect(line)
            widths.append(br.width())
            heights.append(br.height())
        # if self.direction == Qt.Vertical:
        #     return QSize(max(widths), max(heights))
        # else:
        #     return QSize(max(widths), sum(heights))
        return QSize(max(widths), sum(heights))

    def set_text(self, text: str):
        if text is None:
            self.text_label.clear()
            return

        self.text_label.setText(text)

    def set_icon(self, icon: Union[QIcon, QPixmap]):
        if icon is None:
            self.icon_label.clear()
            return

        if isinstance(icon, QIcon):
            icon = icon.pixmap(self._get_correct_image_size(), QIcon.Normal, QIcon.On)
        elif isinstance(icon, QPixmap):
            icon = icon.scaled(self._get_correct_image_size(), Qt.KeepAspectRatioByExpanding, Qt.SmoothTransformation)
        self.icon_label.setPixmap(icon)

    def sizeHint(self) -> PySide6.QtCore.QSize:
        return self.layout.sizeHint()

    @property
    def layout(self) -> QBoxLayout:
        return super().layout()
# region [Main_Exec]


if __name__ == '__main__':
    app = QApplication()
    w = ImageTextButton(icon=QPixmap(r"D:\Dropbox\hobby\Modding\Programs\Github\My_Repos\GidAppTools\gidapptools\data\images\placeholder.png"), text="this is a test" + ('\n' + "this is a test") * 10, direction=Qt.Vertical)
    w.show()
    sys.exit(app.exec())

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from types import UnionType
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final, Callable,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR: Path = Path(__file__).parent.absolute().resolve()

# endregion [Constants]


class ConsoleCodeTypus(Enum):
    FOREGROUND = auto()
    BACKGROUND = auto()
    MODIFICATION = auto()


INDICATOR_REGEX: re.Pattern = re.compile(r"\<(?P<typus>(SET|UNSET))_CONSOLE_CODE:(?P<names>[\w\|]+)\>")


class ConsoleColor:
    typus: ConsoleCodeTypus = None

    __slots__ = ("_name", "_code")

    def __init__(self, name: str, code: str) -> None:
        self._name = name
        self._code = code

    @property
    def name(self) -> str:
        return self._name

    @property
    def code(self) -> str:
        return self._code

    @property
    def set_indicator(self) -> str:
        return f"<SET_CONSOLE_CODE:{self.name}>"

    @property
    def unset_indicator(self) -> str:
        return f"<UNSET_CONSOLE_CODE:{self.name}>"

    def __call__(self, text: str) -> str:
        return f"{self.set_indicator}{text}{self.unset_indicator}"

    def __or__(self, other: Any) -> UnionType:
        if isinstance(other, ConsoleColor):
            if other.name in self.name.split("|"):
                return self
            else:
                return ConsoleColor(name=self.name + "|" + other.name, code=self.code + other.code)

        return NotImplemented

    def __str__(self) -> str:
        return self.code

    def __hash__(self):
        return hash((self._name, self._code))


class ConsoleForegroundColor(ConsoleColor):
    typus: ConsoleCodeTypus = ConsoleCodeTypus.FOREGROUND


class _ConsoleCodeDescriptor:

    def __init__(self, klass: type, code: str) -> None:
        self._klass = klass
        self._code = code
        self._name = None
        self._obj = None

    def __get__(self, obj: "ConsoleCodes", objtype=None):
        if self._obj is None:
            code = self._code if obj.is_applicable is True else ""
            self._obj = self._klass(name=self._name, code=code)
        return self._obj

    def __set_name__(self, owner, name: str):

        self._name = name.lstrip("_")


class ConsoleCodes:
    """
    ANSI color codes
    based on: https://gist.github.com/rene-d/9e584a7dd2935d0f461904b9f2950007
    """

    # __slots__ = ('_BLACK', '_RED', '_GREEN', '_BROWN', '_BLUE', '_PURPLE', '_CYAN', '_LIGHT_GRAY', '_DARK_GRAY', '_LIGHT_RED', '_LIGHT_GREEN', '_YELLOW', '_LIGHT_BLUE', '_LIGHT_PURPLE', '_LIGHT_CYAN', '_LIGHT_WHITE', '_WHITE',
    #              '_BOLD', '_FAINT', '_ITALIC', '_UNDERLINE', '_BLINK', '_NEGATIVE', '_CROSSED',
    #              '_RESET')

    BLACK = _ConsoleCodeDescriptor(ConsoleForegroundColor, code="\033[0;30m")
    RED = _ConsoleCodeDescriptor(ConsoleForegroundColor, code="\033[0;31m")

    def __init__(self):
        self._GREEN = ConsoleForegroundColor(name="GREEN", code="\033[0;32m" if self.is_applicable else "")
        self._BROWN = ConsoleForegroundColor(name="BROWN", code="\033[0;33m" if self.is_applicable else "")
        self._BLUE = ConsoleForegroundColor(name="BLUE", code="\033[0;34m" if self.is_applicable else "")
        self._PURPLE = ConsoleForegroundColor(name="PURPLE", code="\033[0;35m" if self.is_applicable else "")
        self._CYAN = ConsoleForegroundColor(name="CYAN", code="\033[0;36m" if self.is_applicable else "")
        self._LIGHT_GRAY = ConsoleForegroundColor(name="LIGHT_GRAY", code="\033[0;37m" if self.is_applicable else "")
        self._DARK_GRAY = ConsoleForegroundColor(name="DARK_GRAY", code="\033[1;30m" if self.is_applicable else "")
        self._LIGHT_RED = ConsoleForegroundColor(name="LIGHT_RED", code="\033[1;31m" if self.is_applicable else "")
        self._LIGHT_GREEN = ConsoleForegroundColor(name="LIGHT_GREEN", code="\033[1;32m" if self.is_applicable else "")
        self._YELLOW = ConsoleForegroundColor(name="YELLOW", code="\033[1;33m" if self.is_applicable else "")
        self._LIGHT_BLUE = ConsoleForegroundColor(name="LIGHT_BLUE", code="\033[1;34m" if self.is_applicable else "")
        self._LIGHT_PURPLE = ConsoleForegroundColor(name="LIGHT_PURPLE", code="\033[1;35m" if self.is_applicable else "")
        self._LIGHT_CYAN = ConsoleForegroundColor(name="LIGHT_CYAN", code="\033[1;36m" if self.is_applicable else "")
        self._LIGHT_WHITE = ConsoleForegroundColor(name="LIGHT_WHITE", code="\033[1;37m" if self.is_applicable else "")
        self._WHITE = ConsoleForegroundColor(name="WHITE", code="\033[97m" if self.is_applicable else "")

        self._BOLD = "\033[1m" if self.is_applicable else ""
        self._FAINT = "\033[2m" if self.is_applicable else ""
        self._ITALIC = "\033[3m" if self.is_applicable else ""
        self._UNDERLINE = "\033[4m" if self.is_applicable else ""
        self._BLINK = "\033[5m" if self.is_applicable else ""
        self._NEGATIVE = "\033[7m" if self.is_applicable else ""
        self._CROSSED = "\033[9m" if self.is_applicable else ""

        self._RESET = "\033[0m" if self.is_applicable else ""

    @property
    def is_atty(self) -> bool:
        return sys.stdout.isatty()

    @property
    def is_windows(self) -> bool:
        return platform.system().casefold() == "windows"

    @property
    def is_applicable(self) -> bool:
        return all([self.is_atty, self.is_windows])

    @property
    def GREEN(self):
        return self._GREEN

    @property
    def BROWN(self):
        return self._BROWN

    @property
    def BLUE(self):
        return self._BLUE

    @property
    def PURPLE(self):
        return self._PURPLE

    @property
    def CYAN(self):
        return self._CYAN

    @property
    def LIGHT_GRAY(self):
        return self._LIGHT_GRAY

    @property
    def DARK_GRAY(self):
        return self._DARK_GRAY

    @property
    def LIGHT_RED(self):
        return self._LIGHT_RED

    @property
    def LIGHT_GREEN(self):
        return self._LIGHT_GREEN

    @property
    def YELLOW(self):
        return self._YELLOW

    @property
    def LIGHT_BLUE(self):
        return self._LIGHT_BLUE

    @property
    def LIGHT_PURPLE(self):
        return self._LIGHT_PURPLE

    @property
    def LIGHT_CYAN(self):
        return self._LIGHT_CYAN

    @property
    def LIGHT_WHITE(self):
        return self._LIGHT_WHITE

    @property
    def WHITE(self):
        return self._WHITE

    @property
    def BOLD(self):
        return self._BOLD

    @property
    def FAINT(self):
        return self._FAINT

    @property
    def ITALIC(self):
        return self._ITALIC

    @property
    def UNDERLINE(self):
        return self._UNDERLINE

    @property
    def BLINK(self):
        return self._BLINK

    @property
    def NEGATIVE(self):
        return self._NEGATIVE

    @property
    def CROSSED(self):
        return self._CROSSED

    @property
    def RESET(self):
        return self._RESET

    def resolve_text(self, in_text: str) -> str:

        stack = []

        def _replace_func(in_match: re.Match):
            typus = in_match.group("typus")

            names = in_match.group("names").split("|")

            if typus == "SET":
                code = "".join(getattr(self, name).code for name in names)
                stack.append(code)
                return code

            if typus == "UNSET":
                text = self._RESET
                stack.pop(-1)
                if stack:
                    text += stack[-1]
                return text

        return INDICATOR_REGEX.sub(_replace_func, in_text) + self.RESET

# region [Main_Exec]


if __name__ == '__main__':
    col = ConsoleCodes()

    x = col.RED("I am red " + col.BLACK("I am purple") + " I am red")

    print(col.resolve_text(x))

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import platform
from enum import Enum, auto
from typing import Any
from pathlib import Path

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BaseGidEnum(Enum):
    @classmethod
    def _missing_(cls, value: object) -> Any:
        if isinstance(value, str):
            mod_value = value.casefold()
            for member_name, member_value in cls.__members__.items():
                if member_name.casefold() == mod_value or member_value == mod_value:
                    return cls(member_value)
                if isinstance(member_value, str) and member_value.casefold() == mod_value:
                    return cls(member_value)
        return super()._missing_(value)

    @classmethod
    def is_in_value(cls, other: Any) -> bool:
        return other in {member.value for name, member in cls.__members__.items()}

    def __str__(self) -> str:
        return self.name


class OperatingSystem(BaseGidEnum):
    WINDOWS = auto()
    LINUX = auto()
    MAC_OS = auto()
    JYTHON = auto()
    UNDETERMINED = auto()

    @classmethod
    @property
    def default_member(cls) -> "OperatingSystem":
        return cls.UNDETERMINED

    @classmethod
    @property
    def member_str_map(cls) -> dict[str, "OperatingSystem"]:
        return {'windows': cls.WINDOWS,
                'linux': cls.LINUX,
                'darwin': cls.MAC_OS,
                'java': cls.JYTHON}

    @classmethod
    def str_to_member(cls, os_string: str) -> "OperatingSystem":
        def _normalize_name(in_name: str) -> str:
            mod_name = in_name.casefold()
            mod_name = mod_name.strip()
            return mod_name

        return cls.member_str_map.get(_normalize_name(os_string), cls.default_member)

    @classmethod
    def determine_operating_system(cls) -> "OperatingSystem":
        os_string = platform.system()

        return cls.str_to_member(os_string)

    def __str__(self) -> str:
        return self.name


class NamedMetaPath(BaseGidEnum):
    DATA = 'user_data_dir'
    LOG = 'user_log_dir'
    CACHE = 'user_cache_dir'
    CONFIG = 'user_config_dir'
    CONFIG_SPEC = 'user_config_spec_dir'
    STATE = 'user_state_dir'
    SITE_DATA = 'site_data_dir'
    SITE_CONFIG = 'site_config_dir'
    TEMP = 'user_temp_dir'
    DB = 'database_dir'
    DEBUG_DUMP = 'debug_dump_dir'
    _ONLY_FOR_TESTING = '_ONLY_FOR_TESTING'

    def __str__(self) -> str:
        return self.name


class EnvName(str, Enum):
    APP_NAME = 'APP_NAME'
    APP_AUTHOR = 'APP_AUTHOR'
    LOG_DIR = 'APP_LOG_DIR'


# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import shutil
import logging
from pprint import pformat
from typing import Any, Union, TypeVar, Callable, Optional
from pathlib import Path
from threading import Lock, RLock
from tempfile import mkdtemp
from contextlib import contextmanager

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import NotImplementedMetaPath, UnknownMetaPathIdentifier
from gidapptools.utility.enums import NamedMetaPath
from gidapptools.utility.helper import make_pretty
from gidapptools.abstract_classes.abstract_meta_item import AbstractMetaItem

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
log = logging.getLogger(__name__)
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

_T = TypeVar("_T")


class TempPathStorage(set):

    @property
    def discard_lock(self) -> RLock:
        try:
            return self._discard_lock
        except AttributeError:
            self._discard_lock = RLock()
            return self._discard_lock

    def discard_non_existing(self) -> None:
        with self.discard_lock:
            to_discard = [i for i in self if i.exists() is False]

            for item in list(to_discard):
                self.discard(item)

    def add(self, item: _T) -> None:
        self.discard_non_existing()
        return super().add(item)


class MetaPaths(AbstractMetaItem):

    def __init__(self, code_base_dir: Path, paths: dict[NamedMetaPath, Path]) -> None:
        self.code_base_dir = Path(code_base_dir).resolve()
        self._paths: dict[Union[NamedMetaPath, str], Optional[Path]] = paths
        self._created_normal_paths: set[Path] = set()
        self._created_temp_dirs: TempPathStorage[Path] = TempPathStorage()

    def get_path(self, identifier: Union[NamedMetaPath, str], default: Any = NotImplemented) -> Path:
        if isinstance(identifier, str):
            try:
                identifier = NamedMetaPath(identifier)
            except ValueError as e:
                raise UnknownMetaPathIdentifier(identifier=identifier) from e
        path = self._paths.get(identifier, None)
        if path is None:
            if default is NotImplemented:
                raise NotImplementedMetaPath(identifier=identifier)

            path = Path(default)

        if path.exists() is False:
            path.mkdir(parents=True, exist_ok=True)
            self._created_normal_paths.add(path)
        return path

    @property
    def running_pid_storage_file(self) -> Path:
        return self.get_path(NamedMetaPath.DATA).joinpath(".running_instance")

    @property
    def data_dir(self) -> Path:
        return self.get_path(NamedMetaPath.DATA)

    @property
    def cache_dir(self) -> Path:
        return self.get_path(NamedMetaPath.CACHE)

    @property
    def temp_dir(self) -> Path:
        return self.get_path(NamedMetaPath.TEMP)

    @property
    def log_dir(self) -> Path:
        return self.get_path(NamedMetaPath.LOG)

    @property
    def config_dir(self) -> Path:
        return self.get_path(NamedMetaPath.CONFIG)

    @property
    def config_spec_dir(self) -> Path:
        return self.get_path(NamedMetaPath.CONFIG_SPEC)

    @property
    def db_dir(self) -> Path:
        return self.get_path(NamedMetaPath.DB)

    @property
    def debug_dump_dir(self) -> Path:
        return self.get_path(NamedMetaPath.DEBUG_DUMP)

    def _generate_named_temp_dir_path(self, name: str, suffix: str = None, number: int = 0) -> Path:
        _suffix = f"_{suffix}" if suffix is not None else ""
        _number = f"_{number}" if number > 0 else ""
        full_name = f"{name}{_suffix}{_number}"
        return self.temp_dir.joinpath(full_name)

    def get_new_temp_dir(self, suffix: str = None, name: str = None, exists_ok: bool = False) -> Path:
        if name is not None:
            _number = 0
            temp_dir = self._generate_named_temp_dir_path(name, suffix=suffix, number=_number)
            if exists_ok is False:
                while temp_dir.exists():
                    _number += 1
                    temp_dir = self._generate_named_temp_dir_path(name, suffix=suffix, number=_number)

        else:
            temp_dir = Path(mkdtemp(dir=self.temp_dir, suffix=suffix))

        temp_dir.mkdir(parents=True, exist_ok=True)
        self._created_temp_dirs.add(temp_dir)
        return temp_dir

    @contextmanager
    def context_new_temp_dir(self, suffix: str = None, name: str = None, exists_ok: bool = False):
        temp_dir = self.get_new_temp_dir(suffix=suffix, name=name, exists_ok=exists_ok)
        try:
            yield temp_dir
        finally:
            shutil.rmtree(temp_dir)

    def clean_all_temp(self) -> None:
        self._created_temp_dirs.discard_non_existing()
        while len(self._created_temp_dirs) != 0:
            temp_dir = self._created_temp_dirs.pop()

            shutil.rmtree(temp_dir)

    def as_dict(self, pretty: bool = False) -> dict[str, Any]:

        _out = vars(self)
        if pretty is True:
            _out = make_pretty(self)

        return pformat(_out)

    def to_storager(self, storager: Callable = None) -> None:
        if storager is None:
            return
        storager(self)

    def clean_up(self, remove_all_paths: bool = False, **kwargs) -> None:
        self.clean_all_temp()
        if remove_all_paths is True:
            for path in self._created_normal_paths:
                if path.exists():
                    if kwargs.get('dry_run', False) is True:
                        log.debug("Simulating deleting of temp-folder %r and its contents.", path.as_posix())
                    else:
                        log.info("Deleting temp-folder %r and its contents.", path.as_posix())
                        shutil.rmtree(path)

        self.running_pid_storage_file.unlink(missing_ok=True)
        # TODO: find a way to clean shit up, but completely, also add optional kwarg that also removes the author folder


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from warnings import warn, warn_explicit
from functools import wraps

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def mark_experimental(extra_message: str = ""):
    def _wrapper(func):
        @wraps(func)
        def _wrapped(*args, **kwargs):
            func_name = func.__qualname__ or func.__name__
            message = f"The function {func_name!r} is still experimentel and there is no guarantee that it works or does what it says.{extra_message}"
            try:
                warn_explicit(message=message, category=UserWarning, filename=func.__code__.co_filename, lineno=func.__code__.co_firstlineno, module=func.__module__)
            except Exception:
                warn(message=message, category=UserWarning, stacklevel=4)
            return func(*args, **kwargs)

        return _wrapped
    return _wrapper

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QGridLayout

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class JsonEditorWidget(QWidget):

    def __init__(self, parent: Optional[PySide6.QtWidgets.QWidget] = None, f: PySide6.QtCore.Qt.WindowFlags = None) -> None:
        super().__init__(parent, f or Qt.WindowFlags())
        self.setLayout(QGridLayout())

    def layout(self) -> QGridLayout:
        return super().layout()


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from time import perf_counter_ns
from typing import Union, Callable
from pathlib import Path
from functools import wraps, partial
from threading import RLock
from contextlib import contextmanager

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.conversion import seconds2human

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


TIME_NS_FUNC_TYPE = Callable[[], float]


time_execution_path_locks: dict[Path, RLock] = {}


def get_time_execution_path_lock(path: Path) -> RLock:
    lock = time_execution_path_locks.get(path, None)
    if lock is None:
        lock = RLock()
        time_execution_path_locks[path] = lock
    return lock


@contextmanager
def time_execution(identifier: str = None,
                   time_ns_func: TIME_NS_FUNC_TYPE = perf_counter_ns,
                   output: Union[Callable, Path] = partial(print, flush=True),
                   output_kwargs: dict[str, object] = None,
                   condition: Union[bool, Callable[[], bool]] = True,
                   as_seconds: bool = True,
                   decimal_places: Union[int, None] = None,
                   also_pretty: bool = False) -> None:
    if callable(condition):
        condition = condition()
    if condition is True:
        start_time = time_ns_func()
        yield
        end_time = time_ns_func()
        full_time = end_time - start_time
        if as_seconds is True:
            from gidapptools.general_helper.conversion import ns_to_s
            full_time = ns_to_s(full_time, decimal_places=decimal_places)
            unit = 's'
            pretty = "" if also_pretty is False else f" ({seconds2human(full_time)})"
        else:
            unit = 'ns'
            pretty = "" if also_pretty is False else f" ({seconds2human(ns_to_s(full_time))})"
        identifier = 'time_execution' if identifier is None else identifier

        if isinstance(output, Path):
            with get_time_execution_path_lock(output):
                with output.open('a', encoding='utf-8', errors='ignore') as f:
                    f.write(f"{identifier} took {full_time:f} {unit}{pretty}" + '\n')
        else:
            output_kwargs = output_kwargs or {}
            output(f"{identifier} took {full_time:f} {unit}{pretty}", **output_kwargs)
    else:
        yield


def time_func(time_ns_func: TIME_NS_FUNC_TYPE = perf_counter_ns,
              output: Callable = partial(print, flush=True),
              output_kwargs: dict[str, object] = None,
              use_qualname: bool = True,
              condition: Union[bool, Callable[[], bool]] = True,
              as_seconds: bool = True,
              decimal_places: int = None,
              also_pretty: bool = False):

    def _decorator(func):
        func_name = func.__name__ if use_qualname is False else func.__qualname__
        if callable(condition):
            _actual_condition = condition()
        else:
            _actual_condition = condition

        @wraps(func)
        def _wrapped(*args, **kwargs):
            with time_execution(f"executing {func_name!r}", time_ns_func=time_ns_func, output=output, as_seconds=as_seconds, decimal_places=decimal_places, condition=True, also_pretty=also_pretty, output_kwargs=output_kwargs):
                return func(*args, **kwargs)

        if _actual_condition:
            return _wrapped
        return func

    return _decorator


# def profile(func):
#     """
#     from `https://gist.github.com/pavelpatrin/5a28311061bf7ac55cdd`
#     """

#     @wraps(func)
#     def wrapper(*args, **kwargs):
#         from line_profiler import LineProfiler
#         prof = LineProfiler()
#         try:
#             return prof(func)(*args, **kwargs)
#         finally:
#             prof.print_stats()

#     return wrapper


def profile(func):
    """
    Dummy decorator to be able to leave the `line_profiler`-decorator `` in place,
    even when not line-profiling.
    """
    return func


def get_dummy_profile_decorator_in_globals():

    if os.getenv("LINE_PROFILE_RUNNING", "0") == "1" or __debug__ is not True:
        return

    if not isinstance(__builtins__, dict) or 'profile' not in __builtins__:
        __builtins__["profile"] = profile
    # stk = inspect.stack()[1]
    # mod = inspect.getmodule(stk[0])
    # if mod is not None:
    #     setattr(mod, "profile", profile)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
from typing import Optional
from pathlib import Path
from collections import deque

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QFont, QColor, QTextCharFormat, QSyntaxHighlighter
from PySide6.QtCore import Signal, QObject
from PySide6.QtWidgets import QWidget, QTextEdit, QGridLayout

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class StreamCaptureSignaler(QObject):
    flushed_text = Signal(str)
    flushed_items = Signal(tuple)


class BaseStdStreamCapturer:
    signaler: "StreamCaptureSignaler" = None
    separator: str = '\n' + ("═" * 150) + '\n'

    def __init__(self, pass_through: bool = True) -> None:
        self.pass_through = pass_through
        if self.__class__.signaler is None:
            self.__class__.signaler = StreamCaptureSignaler()
        self.original_stream = None
        self.text_cache: str = ""
        self.text_items: list[str] = []

    @property
    def flushed_text(self) -> Signal:
        return self.signaler.flushed_text

    @property
    def flushed_items(self) -> Signal:
        return self.signaler.flushed_items

    def isatty(self) -> bool:
        return False

    def write(self, text: str):
        self.text_cache += text
        if self.original_stream and self.pass_through:
            self.original_stream.write(text)

    def flush(self):
        self.text_items.append(self.text_cache.rstrip() + self.separator)

        self.text_cache = ""
        if self.signaler:
            try:
                self.signaler.flushed_items.emit(self.get_items())
                self.signaler.flushed_text.emit(self.get_text())
            except RuntimeError:
                pass
        if self.original_stream and self.pass_through:
            self.original_stream.flush()

    def get_text(self) -> str:
        return ''.join(self.text_items)

    def get_items(self) -> tuple[str]:
        return tuple(self.text_items)

    def clear(self):
        self.text_items.clear()


class LimitedStdStreamCapturer(BaseStdStreamCapturer):

    def __init__(self, max_len: int = 50) -> None:
        super().__init__()
        self.text_items = deque(maxlen=max_len)


class ErrorHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.rules: dict[re.Pattern, QTextCharFormat] = {}
        self.setup_rules()

    def setup_rules(self):
        sep_pattern = re.compile(r"═+")
        sep_format = QTextCharFormat()
        sep_format.setFontWeight(100)
        sep_format.setBackground(QColor(150, 150, 150, 50))
        self.rules[sep_pattern] = sep_format

        string_pattern = re.compile(r'(?<=\").*(?=\")')
        string_format = QTextCharFormat()
        string_format.setFontItalic(True)
        string_format.setFontUnderline(True)
        string_format.setForeground(QColor(0, 150, 0, 255))
        self.rules[string_pattern] = string_format

        quotes_pattern = re.compile(r'\"')
        quotes_format = QTextCharFormat()
        quotes_format.setForeground(QColor(150, 150, 0, 255))
        self.rules[quotes_pattern] = quotes_format

        self_pattern = re.compile(r"(?<=\b)self(?=\b)")
        self_format = QTextCharFormat()
        self_format.setForeground(QColor(255, 100, 0, 255))
        self.rules[self_pattern] = self_format

        method_pattern = re.compile(r"(?<=\.)\w+(?=\()")
        method_format = QTextCharFormat()
        method_format.setForeground(QColor(0, 0, 200, 255))
        method_format.setFontItalic(True)
        self.rules[method_pattern] = method_format

        error_name_pattern = re.compile(r"\w+Error")
        error_name_format = QTextCharFormat()
        error_name_format.setForeground(QColor(150, 0, 0, 255))
        error_name_format.setTextOutline(QColor(0, 0, 0, 100))
        error_name_format.setFontWeight(1000)
        error_name_format.setFontUnderline(True)
        self.rules[error_name_pattern] = error_name_format

        traceback_title_pattern = re.compile(re.escape(r"Traceback (most recent call last):"))
        traceback_title_format = QTextCharFormat()
        traceback_title_format.setFontUnderline(True)
        traceback_title_format.setBackground(QColor(200, 200, 0, 50))
        self.rules[traceback_title_pattern] = traceback_title_format

    def highlightBlock(self, text: str) -> None:

        for pattern, text_format in self.rules.items():
            for match in pattern.finditer(text):

                start, end = match.span()
                self.setFormat(start, end - start, text_format)


class StdStreamWidget(QWidget):

    def __init__(self, stream_capturer: "BaseStdStreamCapturer", parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.stream_capturer = stream_capturer

        self.setLayout(QGridLayout())
        self.text_view: QTextEdit = None

    def setup(self) -> "StdStreamWidget":
        self.resize(1500, 500)
        self.highlighter = ErrorHighlighter()
        self.text_view = QTextEdit(self)
        self.highlighter.setDocument(self.text_view.document())
        self.text_view.setReadOnly(True)
        self.text_view.setLineWrapMode(QTextEdit.NoWrap)
        font: QFont = self.text_view.font()
        font.setStyleHint(QFont.Monospace)
        font.setFamily("Consolas")
        font.setPointSize(11)
        self.text_view.setFont(font)
        self.layout.addWidget(self.text_view)
        self.text_view.setPlainText(self.stream_capturer.get_text())
        self.stream_capturer.flushed_text.connect(self.set_text)

        return self

    def set_text(self, text: str):
        self.text_view.setPlainText(text)
        self.text_view.verticalScrollBar().setValue(self.text_view.verticalScrollBar().maximum())

    @property
    def layout(self) -> QGridLayout:
        return super().layout()

    def show(self) -> None:
        self.text_view.verticalScrollBar().setValue(self.text_view.verticalScrollBar().maximum())
        return super().show()


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import json
from typing import TYPE_CHECKING, Union, Literal, Callable, Optional
from types import ModuleType
from pathlib import Path
from threading import RLock

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import SpecDataMissingError
from gidapptools.custom_types import PATH_TYPE
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals
from gidapptools.general_helper.class_helper import MethodEnabledWeakSet
from gidapptools.gid_config.conversion.spec_item import SpecEntry, SpecSection
from gidapptools.general_helper.mixins.file_mixin import FileMixin
from gidapptools.gid_config.conversion.converter_grammar import ConverterSpecData, parse_specification

import orjson


# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    pass

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class SpecLoader:
    __slots__ = ("spec_section_class",
                 "spec_entry_class",
                 "parse_func",
                 "loaded_data")

    def __init__(self,
                 spec_section_class: type[SpecSection] = SpecSection,
                 spec_entry_class: type[SpecEntry] = SpecEntry,
                 parse_function: Callable[[str], ConverterSpecData] = parse_specification) -> None:
        self.spec_section_class = spec_section_class
        self.spec_entry_class = spec_entry_class
        self.parse_func = parse_function
        self.loaded_data: dict[str, SpecSection] = None

    @property
    def is_loaded(self) -> bool:
        return self.loaded_data is not None

    def process_entry(self, entry_name: str, entry_data: dict[str, object]) -> SpecEntry:
        entry_data = dict(entry_data)
        if entry_data.get("converter", None) is not None:
            entry_data["converter"] = self.parse_func(entry_data["converter"])
        return self.spec_entry_class(name=entry_name, **entry_data)

    def process_section(self, section_name: str, section_data: dict[str, object]) -> SpecSection:
        section_data = {k: v for k, v in section_data.items() if k != "entries"}
        if section_data.get("default_converter", None) is not None:
            section_data["default_converter"] = self.parse_func(section_data["default_converter"])

        return self.spec_section_class(name=section_name, **section_data)

    def process_data(self, data: dict[str, object]) -> None:
        self.reset()
        sections_data = data["sections"]
        for section_name, section_data in sections_data.items():
            section = self.process_section(section_name=section_name, section_data=section_data)

            for entry_name, entry_data in section_data["entries"].items():
                section.add_entry(self.process_entry(entry_name=entry_name, entry_data=entry_data))
            self.loaded_data[section_name] = section

    def create_dynamic_entry(self, section: SpecSection, entry_name: str) -> SpecEntry:
        entry = self.spec_entry_class(name=entry_name)
        entry.set_section(section)
        return entry

    def reset(self) -> None:
        self.loaded_data = {}

    def __repr__(self) -> str:

        return f'{self.__class__.__name__}(spec_section_class={self.spec_section_class!r}, spec_entry_class={self.spec_entry_class!r})'


class SpecData:

    def __init__(self, name: str, loader: SpecLoader) -> None:
        self.name = name
        self.loader = loader
        self.load_lock: RLock = RLock()
        self._original_data: dict[str, object] = None

    @property
    def original_data(self) -> Optional[dict[str, object]]:
        with self.load_lock:
            return self._original_data

    @property
    def sections(self) -> dict[str, SpecSection]:
        with self.load_lock:
            return {k: v for k, v in self.loader.loaded_data.items()}

    @property
    def entries(self) -> dict[tuple[str, str], SpecEntry]:
        with self.load_lock:
            _out = {}
            for section in self.loader.loaded_data.values():
                for entry in section.entries.values():
                    _out[(section.name, entry.name)] = entry
        return _out

    def get_spec_entry(self, section_name: str, entry_name: str) -> SpecEntry:
        with self.load_lock:
            section = self.sections[section_name]
            try:
                return section[entry_name]
            except KeyError as e:
                if section.dynamic_entries_allowed is True and entry_name is not None:
                    return self.loader.create_dynamic_entry(section=section, entry_name=entry_name)
                else:
                    raise SpecDataMissingError(f"No entry with name {entry_name!r} in section {section_name!r} of {self!r}.") from e

    def load_data(self, data: dict) -> "SpecData":
        with self.load_lock:
            self._original_data = data
            self.loader.process_data(self.original_data)
        return self

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}(name={self.name!r}, loader={self.loader!r})'


class SpecFile(FileMixin, SpecData):
    _name_suffixes_to_remove: tuple[str] = ("_configspec", "_spec", "_config", "_config_spec", "_spec_config", "_specconfig")

    json_load_func: Callable[[str], dict[str, object]] = orjson.loads

    def __init__(self,
                 file_path: PATH_TYPE,
                 loader: SpecLoader = None,
                 changed_parameter: Union[Literal['size'], Literal['file_hash']] = 'size') -> None:
        super().__init__(name=self._generate_name_from_path(file_path, suffixes_to_remove=self._name_suffixes_to_remove),
                         loader=loader or SpecLoader(),
                         file_path=file_path,
                         changed_parameter=changed_parameter,
                         missing_ok=False)
        self._on_reload_targets: MethodEnabledWeakSet = MethodEnabledWeakSet()

    @property
    def original_data(self) -> dict:
        self.reload_if_changed()
        return self._original_data

    @property
    def sections(self) -> dict[str, SpecSection]:
        self.reload_if_changed()
        return super().sections

    @property
    def entries(self) -> dict[tuple[str, str], SpecEntry]:
        self.reload_if_changed()
        return super().entries

    def add_on_reload_target(self, target: Callable[[], None]) -> None:
        self._on_reload_targets.add(target)

    def reload_if_changed(self) -> None:
        if self._original_data is None or self.has_changed is True:
            self.reload()

    def reload(self) -> "SpecFile":
        self.load()
        for target in self._on_reload_targets:
            target()
        return self

    def load(self) -> "SpecFile":
        with self.lock:

            self.load_data(self.json_load_func(self.read()))

        return self

    def save(self) -> None:
        with self.lock:
            data = self.original_data
            json_data = json.dumps(data, indent=4, sort_keys=False)
            self.write(json_data)

    def __repr__(self) -> str:

        return f'{self.__class__.__name__}(name={self.name!r}, file_path={self.file_path.as_posix()!r})'


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from pathlib import Path

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def cli_show_info(argv=None):
    argv = argv or sys.argv[1:]
# region [Main_Exec]


if __name__ == '__main__':
    cli_show_info()

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
import sys
import logging
from typing import TYPE_CHECKING, Any
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.logger import get_logger, get_meta_logger

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_logger.records import LOG_RECORD_TYPES

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ProhibitiveSingletonMeta(type):
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is not None:
            raise RuntimeError(f"There can only be one instance of {cls.__name__}")
        cls._instance = super(ProhibitiveSingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instance


class QtMessageHandler(metaclass=ProhibitiveSingletonMeta):
    received_records: list["LOG_RECORD_TYPES"] = []
    __slots__ = ("msg_split_regex",
                 "is_installed",
                 "_old_messagehandler")

    def __init__(self) -> None:
        self.msg_split_regex = re.compile(r"(?P<q_class>.*)\:\:(?P<q_method>.*)\:(?P<actual_message>.*)")
        self.is_installed: bool = False
        self._old_messagehandler = None

    def install(self, overwrite_install: bool = False) -> "QtMessageHandler":
        if self.is_installed is True and overwrite_install is False:
            return self
        from PySide6.QtCore import qInstallMessageHandler
        self._old_messagehandler = qInstallMessageHandler(self)
        self.is_installed = True

        return self

    def mode_to_log_level(self, in_mode):
        in_mode = str(in_mode).rsplit('.', maxsplit=1)[-1].strip().removeprefix("Qt").removesuffix("Msg").upper()
        if in_mode == "FATAL":
            in_mode = "ERROR"
        elif in_mode == "SYSTEM":
            in_mode = "INFO"
        return logging.getLevelName(in_mode)

    def get_context(self, in_context: None):
        _logger = None
        frame_count = 2
        while _logger is None:
            try:
                frame = sys._getframe(frame_count)
                if frame is None:
                    raise RuntimeError(f"unable to get a frame for {in_context!r}")

                _context_data = {"fn": in_context.file or frame.f_code.co_filename,
                                 "func": in_context.function or frame.f_code.co_name,
                                 "lno": in_context.line or frame.f_lineno}

                _logger = get_logger(frame.f_globals["__name__"])
            except AttributeError:
                frame_count = frame_count + 1
        return _context_data, _logger

    def modify_message(self, in_msg: str) -> str:

        if re_match := self.msg_split_regex.match(in_msg):
            named_parts = re_match.groupdict()
            _message = named_parts.pop("actual_message").strip()
            return _message, {"is_qt": True} | named_parts

        return in_msg, {"is_qt": True}

    def __call__(self, mode, context, message) -> Any:
        try:
            context_data, logger = self.get_context(context)
        except Exception as e:
            logger = get_meta_logger()
            logger.error("encountered Error %r while trying to log QT-Message %r", e, message, exc_info=True)
            return
        log_level = self.mode_to_log_level(mode)
        msg, extras = self.modify_message(message)
        record = logger.makeRecord(logger.name, log_level, msg=msg, extra=extras, exc_info=None, args=None, ** context_data)
        logger.handle(record)
        self.received_records.append(record)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, auto
from pathlib import Path

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]
SPEC_BASE_DATA = {
    "__env__": {
        "__default__": {"converter": "string"},
        "PATH": {"converter": "list(sub_typus=path, split_char=;)"},
        "APPDATA": {"converter": "path"},
        "LOCALAPPDATA": {"converter": "path"},
        "PROGRAMFILES(X86)": {"converter": "path"},
        "PROGRAMFILES": {"converter": "path"},
        "PROGRAMDATA": {"converter": "path"},
        "PATHEXT": {"converter": "list(sub_typus=string, split_char=;)"},
        "USERPROFILE": {"converter": "path"},
        "IS_DEV": {"converter": "boolean",
                   "default": "false"},
        "TMP": {"converter": "path"},
        "TEMP": {"converter": "path"},
        "WINDIR": {"converter": "path"}
    }
}


class NonTypeBaseTypus(Enum):
    DEFAULT = auto()
    FILE_SIZE = auto()
    STRING_CHOICE = auto()


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from typing import Union
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def amount_lines_in_file(in_file: PATH_TYPE):
    in_file = Path(in_file)
    if in_file.is_file() is False:
        raise FileNotFoundError(f"The path {in_file.as_posix()!r} is not a file.")
    with in_file.open("r", encoding='utf-8', errors='ignore') as f:

        count = sum(1 for _ in f)

    return count


def get_last_line(in_file: PATH_TYPE, decode: bool = True, encoding: str = "utf-8", errors: str = "ignore") -> str:
    in_file = Path(in_file)
    if in_file.is_file() is False:
        raise FileNotFoundError(f"The path {in_file.as_posix()!r} is not a file.")
    with in_file.open('rb') as f:
        f.seek(-2, os.SEEK_END)
        while f.read(1) != b'\n':
            f.seek(-2, os.SEEK_CUR)
        last_line = f.readline()
        if last_line == b"\r\n":
            last_line = b""
        if decode is True:
            last_line = last_line.decode(encoding=encoding, errors=errors)
        return last_line


def escalating_find_file(file_name: str, directory: Union[str, os.PathLike], case_sensitive: bool = False) -> Path:

    def _case_sensitive_compare(in_file_name: str, target_file_name: str) -> bool:
        return in_file_name == target_file_name

    def _case_insensitive_compare(in_file_name: str, target_file_name: str) -> bool:
        return in_file_name.casefold() == target_file_name

    directory = Path(directory).resolve()
    if directory.is_dir() is False:
        raise NotADirectoryError(f"The path {directory.as_posix()!r} is not a directory.")

    compare_func = _case_sensitive_compare

    if case_sensitive is False:
        file_name = file_name.casefold()
        compare_func = _case_insensitive_compare
    for file in directory.iterdir():
        if not file.is_file():
            continue
        if compare_func(file.name, file_name) is True:
            return file.resolve()

    if len(directory.parts) <= 1:
        raise FileNotFoundError(f"Unable to find the file with the name {file_name!r}.")

    return escalating_find_file(file_name=file_name, directory=directory.parent, case_sensitive=case_sensitive)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Iterable, Optional, Union, Protocol, Callable, TypeAlias, TypeGuard, TypedDict, NamedTuple, TypeVar, TypeVarTuple, TYPE_CHECKING
from pathlib import Path
import numpy as np
from abc import ABC, abstractmethod

from functools import partial
import random
from copy import copy
from decimal import Decimal
from enum import Enum, auto, Flag
import scipy
from gidapptools.errors import MissingOptionalDependencyError
from gidapptools.gidcolor.misc_calculations import calculate_contrast_ratio
from math import radians, degrees, ceil, floor, sqrt
from functools import total_ordering
import sys


from gidapptools.gidcolor.coversion import hsv_to_rgb, rgb_to_hsv, rgb_int_to_rgb_float, rgb_to_hsl, hsl_to_rgb

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
try:
    from PySide6.QtGui import QColor
    PYSIDE6_AVAILABLE = True
except ImportError:
    PYSIDE6_AVAILABLE = False

if TYPE_CHECKING:
    from PySide6.QtGui import QColor

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

RGB_FLOAT_TO_INT_FACTOR: float = 1.0 / 255
# endregion [Constants]

# region [Types]

_FLOAT_COLOR_VALUE: TypeAlias = tuple[float, float, float]

_FLOAT_COLOR_W_ALPHA_VALUE: TypeAlias = tuple[float, float, float, float]

_INT_COLOR_VALUE: TypeAlias = tuple[int, int, int]

_INT_COLOR_W_ALPHA_VALUE: TypeAlias = tuple[int, int, int, float]

# endregion [Types]


def _clamp_float(in_float: float, minimum: float, maximum: float) -> float:
    return min(maximum, max(minimum, in_float))


def _clamp_between_zero_one(in_float: float) -> float:
    return _clamp_float(in_float=in_float, minimum=0.0, maximum=1.0)


def percent_float_to_degrees(in_value: float) -> float:
    return (in_value * 360) % 360


def degrees_to_percent_float(in_value: float) -> float:
    return (in_value / 360) % 1.0


def truncate_float(value: float, ndigits: int) -> float:
    float_string = str(value)
    pre_comma, post_comma = float_string.split(".")
    return float(pre_comma + '.' + post_comma[:ndigits])


class BasicFormat(Enum):
    DEFAULT = auto()

    RGB = auto()
    HSV = auto()
    HSL = auto()

    RGB_INT = auto()

    HEX = auto()

    CSS = auto()

    @classmethod
    def from_string(cls, in_string: str) -> Self:
        in_string = in_string.casefold()
        for name, member in cls._member_map_.items():
            if name.casefold() == in_string:
                return member

        raise ValueError(f"No Member with name {in_string!r}.")


@total_ordering
class Color:

    light_dark_factor_threshold: float = 127.5

    __slots__ = ("_rgb",
                 "_hsv",
                 "_hsl",
                 "alpha",
                 "_qcolor")

    def __init__(self,
                 rgb: _FLOAT_COLOR_VALUE,
                 hsv: _FLOAT_COLOR_VALUE,
                 hsl: _FLOAT_COLOR_VALUE,
                 alpha: float = 1.0) -> None:

        self._rgb: np.ndarray
        self._hsl: np.ndarray
        self._hsv: np.ndarray

        super().__setattr__("_rgb", np.asfarray([_clamp_between_zero_one(i) for i in rgb], dtype=np.float32))
        super().__setattr__("_hsv", np.asfarray([_clamp_between_zero_one(i) for i in hsv], dtype=np.float32))
        super().__setattr__("_hsl", np.asfarray([_clamp_between_zero_one(i) for i in hsl], dtype=np.float32))

        self.alpha: float

        super().__setattr__("alpha", alpha)
        super().__setattr__("_qcolor", None)

    @property
    def rgb(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        return tuple(self._rgb.tolist())

    @property
    def rgba(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        return tuple(self._rgb.tolist()) + (self.alpha,)

    @property
    def hsv(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        return tuple(self._hsv.tolist())

    @property
    def hsva(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        return tuple(self._hsv.tolist()) + (self.alpha,)

    @property
    def hsl(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        return tuple(self._hsl.tolist())

    @property
    def hsla(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        return tuple(self._hsl.tolist()) + (self.alpha,)

    @property
    def hls(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        hsl = self.hsl
        return (hsl[0], hsl[2], hsl[1])

    @property
    def hlsa(self) -> _FLOAT_COLOR_W_ALPHA_VALUE:
        hsl = self.hsl
        return (hsl[0], hsl[2], hsl[1]) + (self.alpha,)

    @property
    def rgb_int(self) -> _INT_COLOR_VALUE:
        return tuple(round(i * 255) for i in self._rgb)

    @property
    def rgba_int(self) -> _INT_COLOR_W_ALPHA_VALUE:
        return tuple(round(i * 255) for i in self._rgb) + (self.alpha,)

    @property
    def hex(self) -> str:
        return '#' + ''.join(f"{p:02X}" for p in (self.rgb_int + (round(self.alpha * 255),)))

    @property
    def hex_no_alpha(self) -> str:
        return '#' + ''.join(f"{p:02X}" for p in self.rgb_int)

    @property
    def qcolor(self) -> "QColor":
        if PYSIDE6_AVAILABLE is False:
            raise MissingOptionalDependencyError("PySide6", "gidapptools")

        if self._qcolor is None:
            super().__setattr__("_qcolor", QColor.fromRgbF(*self._rgb, a=self._alpha))

        return self._qcolor

    @property
    def hue(self) -> float:
        return float(self.hsv[0])

    @property
    def hue_degrees(self) -> float:
        return round(percent_float_to_degrees(self.hsv[0]), ndigits=3)

    @classmethod
    def from_rgb(cls, r: float, g: float, b: float, alpha: float = 1.0) -> Self:
        rgb_value = (r, g, b)
        return cls(rgb=rgb_value,
                   hsv=rgb_to_hsv(*rgb_value),
                   hsl=rgb_to_hsl(*rgb_value),
                   alpha=alpha)

    @classmethod
    def from_rgb_int(cls, r: int, g: int, b: int, alpha: float = 1.0) -> Self:
        r_float, g_float, b_float = rgb_int_to_rgb_float((r, g, b))
        return cls.from_rgb(r=r_float,
                            g=g_float,
                            b=b_float,
                            alpha=alpha)

    @classmethod
    def from_hsl(cls, h: float, s: float, l: float, alpha: float = 1.0) -> Self:
        v = hsl_to_rgb(h, s, l)

        return cls.from_rgb(*v, alpha=alpha)

    @classmethod
    def from_hsv(cls, h: float, s: float, v: float, alpha: float = 1.0) -> Self:
        return cls.from_rgb(*hsv_to_rgb(h, s, v), alpha=alpha)

    @classmethod
    def from_hex(cls, value: str) -> Self:
        rgb_int_values = tuple(bytes.fromhex(value.removeprefix("#")))
        if len(rgb_int_values) == 4:
            rgb_int_values = (*rgb_int_values[:-1], float(rgb_int_values[-1] / 255))

        return cls.from_rgb_int(*rgb_int_values)

    @classmethod
    def from_qcolor(cls, value: "QColor") -> Self:
        if PYSIDE6_AVAILABLE is False:
            raise MissingOptionalDependencyError("PySide6", "gidapptools")
        instance = cls.from_rgb(*value.getRgbF())
        super(cls, instance).__setattr__("_qcolor", value)
        return instance

    def get_complementary_color(self) -> Self:
        old_hsl = self.hsl
        new_hsl = ((old_hsl[0] + 0.5), old_hsl[1], old_hsl[2])

        return self.__class__.from_hsl(*new_hsl, alpha=self.alpha)

    def contrast_ratio(self, other_color: Self) -> float:
        return calculate_contrast_ratio(self.rgb, other_color.rgb)

    def determine_light_dark_factor(self) -> float:
        r, g, b = self.rgb_int
        return sqrt((0.299 * (r * r)) + (0.587 * (g * g)) + (0.114 * (b * b)))

    @property
    def is_light(self) -> bool:
        return self.determine_light_dark_factor() > self.light_dark_factor_threshold

    @property
    def is_dark(self) -> bool:
        return self.determine_light_dark_factor() < self.light_dark_factor_threshold

    def as_string(self, string_format: BasicFormat = BasicFormat.DEFAULT, with_alpha: bool = True) -> str:

        if string_format in {BasicFormat.RGB, BasicFormat.HSL, BasicFormat.HSV, BasicFormat.RGB_INT}:
            attr_name = string_format.name.casefold()
            if with_alpha is True:
                attr_name += "a"

            return str(getattr(self, attr_name))

        if string_format is BasicFormat.CSS:
            if with_alpha is True:
                return f"rgba{self.rgba_int!s}"
            else:
                return f"rgb{self.rgb_int!s}"

        if string_format is BasicFormat.HEX:
            if with_alpha is True:
                return str(self.hex)
            else:
                return str(self.hex_no_alpha)

        return str(self)

    def as_pillow_color_value(self) -> tuple[int, int, int, int]:
        return self.rgb_int + (round(self.alpha * 255),)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Color):
            return self._rgb == other._rgb and self.alpha == other.alpha

        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Color):
            return sum(self.rgba) < sum(other.rgba)
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._rgb) + hash(self._hsv) + hash(self._hsl) + hash(self.alpha)

    def __setattr__(self, name: str, value: object) -> None:
        raise AttributeError(f"Unable to set {name!r} to {value!r} for {self!r} as instances of {self.__class__.__name__!r} are immutable.")

    def __format__(self, format_spec: str) -> str:
        with_alpha = None
        if format_spec.startswith(("+a", "-a")):
            with_alpha = format_spec[0] == "+"
            format_spec = format_spec[2:]

        elif format_spec.endswith(("+a", "-a")):
            with_alpha = format_spec[-2] == "+"
            format_spec = format_spec[:-2]

        string_format = BasicFormat.from_string(format_spec)

        return self.as_string(*(i for i in (string_format, with_alpha) if i is not None))

    def __repr__(self) -> str:
        # return f'{self.__class__.__name__}(rgb={self.rgb!r}, hsv={self.hsv!r}, hsl={self.hsl!r}, alpha={self.alpha!r})'
        return f'{self.__class__.__name__}({self.hex!r})'


# region [Main_Exec]


if __name__ == '__main__':
    x = Color.from_hex("#556B2F")
    print(type(x.rgb[0]))

    print(x.hue_degrees)
    print(x)


# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from typing import Any, Union, Iterable
from pathlib import Path
from operator import or_
from warnings import warn
from functools import reduce
from importlib.metadata import entry_points

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.data import ENTRY_POINT_SELECT_ARGS
from gidapptools.errors import NotSetupError, NoFactoryFoundError, MetaItemNotFoundError, RegisterAfterSetupError
from gidapptools.custom_types import PATH_TYPE
from gidapptools.meta_data.meta_info import MetaInfo, MetaInfoFactory
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.meta_data.meta_paths import MetaPaths, MetaPathsFactory
from gidapptools.meta_data.config_kwargs import ConfigKwargs
from gidapptools.general_helper.dict_helper import SafeMergeDict
from gidapptools.abstract_classes.abstract_meta_item import AbstractMetaItem
from gidapptools.abstract_classes.abstract_meta_factory import AbstractMetaFactory

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


META_ITEMS_TYPE = Any


class AppMeta:
    factories: list[AbstractMetaFactory] = [MetaInfoFactory,
                                            MetaPathsFactory,
                                            # MetaConfigFactory
                                            ]
    plugin_data: list[dict[str, Any]] = []
    default_to_initialize = [factory.product_name for factory in factories]
    default_base_configuration: dict[str, Any] = SafeMergeDict(raise_on_overwrite=True)
    default_base_configuration: dict[str, Any] = reduce(or_, [default_base_configuration] + [factory.__default_configuration__ for factory in factories])

    def __init__(self) -> None:
        self.is_setup = False
        self.meta_items: dict[str, AbstractMetaItem] = {}
        self._get_plugins()

    @property
    def all_item_names(self) -> list[str]:

        return list(self.meta_items.keys())

    @property
    def all_items(self) -> list[META_ITEMS_TYPE]:
        return list(self.meta_items.values())

    def check_is_setup(self):
        if self.is_setup is False:
            raise NotSetupError(self)

    def _get_plugins(self):
        all_entry_points = entry_points()

        for plugin in all_entry_points.select(**ENTRY_POINT_SELECT_ARGS):

            try:
                loaded_plugin = plugin.load()
                loaded_plugin(self)
            except AttributeError as e:
                warn(f'plugin could not be loaded because of {e}.', stacklevel=4)

    def add_plugin_data(self, factory: AbstractMetaFactory) -> None:
        plugin_dict = {"product_name": factory.product_name,
                       "file": Path(inspect.getfile(factory)).resolve().as_posix(),
                       "module": inspect.getmodule(factory).__name__}
        self.plugin_data.append(plugin_dict)

    def register(self, factory: AbstractMetaFactory, default_configuration: dict[str, Any] = None) -> None:
        if self.is_setup is True:
            raise RegisterAfterSetupError(f'Unable to register new plug-ins after setting up {self.__class__.__name__!r}.')
        if not issubclass(factory, AbstractMetaFactory):
            raise TypeError(f"'factory' needs to be a subclass of {AbstractMetaFactory.__name__!r}.")

        self.factories.append(factory)

        default_configuration = {} if default_configuration is None else default_configuration

        self.default_base_configuration |= factory.__default_configuration__
        self.default_base_configuration |= default_configuration
        self.add_plugin_data(factory=factory)

    def __getitem__(self, item_name) -> META_ITEMS_TYPE:
        self.check_is_setup()
        _out = self.meta_items.get(item_name, MiscEnum.NOTHING)
        if _out is MiscEnum.NOTHING:
            raise MetaItemNotFoundError(item_name, self.all_item_names)
        return _out

    def get(self, item_name: Union[str, type[META_ITEMS_TYPE]] = None) -> META_ITEMS_TYPE:
        if item_name is None:
            self.check_is_setup()
            return dict(self.meta_items)
        if inspect.isclass(item_name):
            item_name = item_name.name
        return self[item_name]

    def __contains__(self, item: Union[str, AbstractMetaItem]) -> bool:
        if isinstance(item, str):
            return item in self.all_item_names
        if isinstance(item, AbstractMetaItem):
            return item in self.all_items
        if isinstance(item, AbstractMetaFactory):
            return item in self.factories
        return NotImplemented

    def _initialize_data(self, config_kwargs: ConfigKwargs) -> None:
        factory_map = {factory.product_name: factory for factory in self.factories}
        for name in config_kwargs.get('items_to_initialize'):
            factory = factory_map.get(name, MiscEnum.NOTHING)
            if factory is MiscEnum.NOTHING:
                raise NoFactoryFoundError(name)
            meta_item = factory.build(config_kwargs)
            self.meta_items[factory.product_name] = meta_item
            config_kwargs.created_meta_items[factory.product_name] = meta_item

    def setup(self, init_path: PATH_TYPE, items_to_initialize: Iterable[str] = None, **kwargs) -> None:
        if self.is_setup is True:
            return
        init_path = Path(init_path)
        items_to_initialize = [] if items_to_initialize is None else items_to_initialize
        items_to_initialize += self.default_to_initialize
        base_configuration = self.default_base_configuration.copy() | {'init_path': init_path, 'items_to_initialize': items_to_initialize}

        config_kwargs = ConfigKwargs(base_configuration=base_configuration, **kwargs)

        self._initialize_data(config_kwargs=config_kwargs)

        self.is_setup = True

    def clean_up(self, **kwargs) -> None:
        """
        possible kwargs:
            remove_all_paths: bool, default=False, remove all paths that were created by meta_paths
            dry_run:bool, default=False, only prints a message, and does not do actuall clean up
        """

        for item in self.all_items:
            try:
                item.clean_up(**kwargs)
            except AttributeError:
                warn(f"Meta-Item {item.name!r} has no Method named 'clean_up'.")

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'


app_meta = AppMeta()


def setup_meta_data(init_path: PATH_TYPE, **kwargs) -> None:
    app_meta.setup(init_path=init_path, **kwargs)


def get_meta_item(item_name: str = None) -> Union[dict[str, type[META_ITEMS_TYPE]], META_ITEMS_TYPE]:
    return app_meta.get(item_name)


def get_meta_info() -> MetaInfo:

    return app_meta['meta_info']


def get_meta_paths() -> MetaPaths:
    return app_meta['meta_paths']


    # region [Main_Exec]
if __name__ == '__main__':
    from faked_pack_src import call_and_return
    call_and_return(setup_meta_data)

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import argparse
from typing import TYPE_CHECKING, Any
from pathlib import Path

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    pass

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class GidBaseCommandLineParser(argparse.ArgumentParser):

    def __init__(self,
                 prog: str = None,
                 fromfile_prefix_chars: str = None,
                 argument_default: Any = None,
                 conflict_handler: str = 'error',
                 add_help: bool = True,
                 allow_abbrev: bool = True,
                 exit_on_error: bool = True) -> None:
        super().__init__(prog=prog,
                         fromfile_prefix_chars=fromfile_prefix_chars,
                         argument_default=argument_default,
                         conflict_handler=conflict_handler,
                         add_help=add_help,
                         allow_abbrev=allow_abbrev,
                         exit_on_error=exit_on_error)


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



import attr


NANOSECONDS_IN_SECOND: int = 1_000_000_000
MICROSECONDS_IN_SECOND: int = 1_000_000

RAW_STRING_TRUE_VALUES: frozenset[str] = frozenset({'yes',
                                                    'y',
                                                    '1',
                                                    'true',
                                                    '+'})

RAW_STRING_FALSE_VALUES: frozenset[str] = frozenset({'no',
                                                     'n',
                                                     '0',
                                                     'false',
                                                     '-'})


STRING_TRUE_VALUES: frozenset[str] = frozenset({str(value).casefold() for value in RAW_STRING_TRUE_VALUES})

STRING_FALSE_VALUES: frozenset[str] = frozenset({str(value).casefold() for value in RAW_STRING_FALSE_VALUES})


FILE_SIZE_SYMBOL_DATA: tuple[tuple[str, str]] = (('K', 'Kilo'),
                                                 ('M', 'Mega'),
                                                 ('G', 'Giga'),
                                                 ('T', 'Tera'),
                                                 ('P', 'Peta'),
                                                 ('E', 'Exa'),
                                                 ('Z', 'Zetta'),
                                                 ('Y', 'Yotta'))


@attr.s(auto_attribs=True, auto_detect=True, frozen=True, slots=True, weakref_slot=True)
class TimeUnit:
    name: str = attr.ib()
    symbol: str = attr.ib()
    factor: float = attr.ib(order=True, eq=True)
    aliases: tuple[str] = attr.ib(converter=tuple, default=tuple())
    plural: str = attr.ib()

    @plural.default
    def default_plural(self):
        return self.name + "s"

    def convert_seconds(self, in_seconds: int) -> int:
        return int(in_seconds / self.factor)

    def convert_with_rest(self, in_seconds: int) -> tuple[int, int]:
        _amount, _rest = divmod(in_seconds, self.factor)

        return int(_amount), _rest

    def value_to_string(self, in_value: int, use_symbols: bool = False) -> str:
        if use_symbols is True:
            return f"{in_value}{self.symbol}"
        if in_value == 1:
            return f"{in_value} {self.name}"
        return f"{in_value} {self.plural}"


RAW_TIMEUNITS: tuple[TimeUnit] = (
    TimeUnit(name='nanosecond', symbol='ns', factor=1 / NANOSECONDS_IN_SECOND, aliases=tuple()),
    TimeUnit(name="microsecond", symbol="us", factor=1 / MICROSECONDS_IN_SECOND, aliases=("mi", "mis", "mü", "müs", "μs")),
    TimeUnit(name='millisecond', symbol='ms', factor=1 / 1000, aliases=tuple()),
    TimeUnit(name='second', symbol='s', factor=1.0, aliases=("sec",)),
    TimeUnit(name='minute', symbol='m', factor=60.0, aliases=("min", "mins")),
    TimeUnit(name='hour', symbol='h', factor=60.0 * 60, aliases=tuple()),
    TimeUnit(name='day', symbol='d', factor=60.0 * 60 * 24, aliases=tuple()),
    TimeUnit(name='week', symbol='w', factor=60.0 * 60 * 24 * 7, aliases=tuple()),
    TimeUnit(name="year", symbol="y", factor=(60.0 * 60 * 24 * 7 * 52) + (60.0 * 60 * 24), aliases=("a",))
)


TIMEUNITS = sorted(RAW_TIMEUNITS, reverse=True)

TIMEUNIT_NAME_MAP = {t.name.casefold(): t for t in TIMEUNITS}



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import sys
import argparse
from typing import IO, TYPE_CHECKING, Any, Union, Literal, TypeVar, Optional, Sequence
from pathlib import Path
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor

# * Third Party Imports --------------------------------------------------------------------------------->
from yarl import URL
from jinja2 import Template, BaseLoader, Environment
from rich.console import Console as RichConsole

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtGui import QIcon, QImage, QPixmap
from PySide6.QtCore import Qt, Slot, QEvent, QObject, QSettings
from PySide6.QtWidgets import QWidget, QMainWindow, QMessageBox, QApplication, QSplashScreen, QSystemTrayIcon

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import NotSetupError, MetaItemNotFoundError, ApplicationExistsError, ApplicationNotSetupError
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.gid_utility.version_item import VersionItem
from gidapptools.general_helper.class_helper import make_repr
from gidapptools.general_helper.string_helper import StringCase, StringCaseConverter
from gidapptools.gidapptools_qt._data.templates import ABOUT_TEMPLATE_FILE, ABOUT_STYLESHEET_FILE, ARG_DOC_HTML_TEMPLATE_FILE, ARG_DOC_MARKDOWN_TEMPLATE_FILE
from gidapptools.gidapptools_qt.basics.sys_tray import GidBaseSysTray
from gidapptools.gidapptools_qt.basics.main_window import GidBaseMainWindow
from gidapptools.meta_data.meta_info.meta_info_item import MetaInfo
from gidapptools.gidapptools_qt.resources.placeholder import QT_DEFAULT_APP_ICON_IMAGE

# * Local Imports --------------------------------------------------------------------------------------->
from gidapptools import get_meta_info

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.resources.resources_helper import PixmapResourceItem

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]-uz1


class AppArgParserResult:

    def __init__(self) -> None:
        self.exec_file: Path = None
        self.main_window_flags = Qt.WindowFlags()
        self.main_window_states = Qt.WindowStates()
        self.show_window: bool = True


class CommandLineArgDoc:

    suppress_indicators: set[str] = {"==SUPPRESS=="}
    single_line_text_template: Template = Environment(loader=BaseLoader).from_string("{{name}} - {{help_text}} - {{argument_strings}} - default: {{default_value}} - required: {{is_required}} - is flag:{{is_flag}}")
    text_template: Template = Environment(loader=BaseLoader).from_string("{{name}}\n{{help_text}}\n{{argument_strings}}\ndefault: {{default_value}}\nrequired: {{is_required}}\nis flag:{{is_flag}}")
    markdown_template: Template = Environment(loader=BaseLoader).from_string(ARG_DOC_MARKDOWN_TEMPLATE_FILE.read_text(encoding='utf-8', errors='ignore'))
    html_template: Template = Environment(loader=BaseLoader).from_string(ARG_DOC_HTML_TEMPLATE_FILE.read_text(encoding='utf-8', errors='ignore'))

    def __init__(self, argument: Union["BaseAppArgParseAction", argparse.Action], app_meta_info: "MetaInfo") -> None:
        self.argument = argument
        self.app_meta_info = app_meta_info

    @property
    def name(self) -> str:
        name = self.argument.metavar or self.argument.dest
        return StringCaseConverter.convert_to(name, StringCase.TITLE)

    @property
    def help_text(self) -> str:
        return self.argument.help.replace("%(prog)r", "{prog}").replace("%(prog)s", "{prog}")

    @property
    def default_value(self) -> Optional[Any]:
        default_value = self.argument.default
        if str(default_value) in self.suppress_indicators:
            return None

        return default_value

    @property
    def is_required(self) -> bool:
        return self.argument.required

    @property
    def is_flag(self) -> bool:
        try:
            return self.argument.is_flag
        except AttributeError:
            return isinstance(self.argument, argparse._StoreAction)

    @property
    def choices(self) -> Optional[Iterable]:
        return self.argument.choices

    @property
    def argument_strings(self) -> tuple[str]:
        return tuple(self.argument.option_strings)

    def get_text(self, single_line: bool = False) -> str:
        bool_values = {True: "Yes", False: "No"}
        help_text = self.help_text.format(prog=self.app_meta_info.pretty_app_name)
        is_required = bool_values[self.is_required]
        is_flag = bool_values[self.is_flag]
        default_value = bool_values.get(self.default_value, self.default_value)

        argument_strings = self.argument_strings
        template = self.single_line_text_template if single_line is True else self.text_template
        return template.render(name=self.name,
                               help_text=help_text,
                               is_required=is_required,
                               is_flag=is_flag,
                               default_value=default_value,
                               argument_strings=argument_strings,
                               prog=self.app_meta_info.pretty_app_name)

    def get_markdown(self) -> str:
        bool_values = {True: "✅", False: "❎"}
        help_text = self.help_text.format(prog="`" + self.app_meta_info.pretty_app_name + "`")
        is_required = bool_values[self.is_required]
        is_flag = bool_values[self.is_flag]
        default_value = bool_values.get(self.default_value, self.default_value)

        argument_strings = '\n'.join(f"- {arg}" for arg in self.argument_strings)

        return self.markdown_template.render(name=self.name,
                                             help_text=help_text,
                                             is_required=is_required,
                                             is_flag=is_flag,
                                             default_value=default_value,
                                             argument_strings=argument_strings,
                                             prog=self.app_meta_info.pretty_app_name)

    def get_html(self) -> str:
        bool_values = {True: '✔️', False: '❌'}
        help_text = self.help_text.format(prog=f'<div class="app_name">{self.app_meta_info.pretty_app_name}</div>')
        is_required = bool_values[self.is_required]
        is_flag = bool_values[self.is_flag]
        default_value = bool_values.get(self.default_value, self.default_value)

        return self.html_template.render(name=self.name,
                                         help_text=help_text,
                                         is_required=is_required,
                                         is_flag=is_flag,
                                         default_value=default_value,
                                         argument_strings=self.argument_strings,
                                         prog=self.app_meta_info.pretty_app_name)

    def __str__(self) -> str:
        return self.get_text(single_line=True)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(action={self.action!r})"


class BaseAppArgParseAction(argparse.Action):

    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 typus=None,
                 choices=None,
                 required=False,
                 helpus=None,
                 metavar=None):
        self.option_strings = option_strings
        self.dest = dest
        self.nargs = nargs
        self.const = const
        self.default = default
        self.type = typus
        self.choices = choices
        self.required = required
        self.help = helpus
        self.metavar = metavar

    def get_doc_item(self, app_meta_info: MetaInfo) -> CommandLineArgDoc:
        return CommandLineArgDoc(self, app_meta_info)

    @property
    def is_flag(self) -> bool:
        return False


class FlagAction(BaseAppArgParseAction):
    def __init__(self,
                 option_strings,
                 dest,
                 default=None,
                 typus=None,
                 choices=None,
                 required=False,
                 helpus=None,
                 metavar=None):

        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

        if helpus is not None and default is not None:
            helpus += " (default: %(default)s)"

        super().__init__(option_strings=_option_strings,
                         dest=dest,
                         nargs=0,
                         default=default,
                         typus=typus,
                         choices=choices,
                         required=required,
                         helpus=helpus,
                         metavar=metavar)

    @property
    def is_flag(self) -> bool:
        return True

    def format_usage(self):
        return ' | '.join(self.option_strings)

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith('--no-'))


class MainWindowMaximizedAction(FlagAction):

    def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace, values: str | Sequence[Any] | None, option_string: str | None = ...) -> None:

        namespace.main_window_states |= Qt.WindowMaximized
        values = Qt.WindowMaximized
        super().__call__(parser=parser, namespace=namespace, values=values, option_string=option_string)


class MainWindowMinimizedAction(FlagAction):

    def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace, values: str | Sequence[Any] | None, option_string: str | None = ...) -> None:

        namespace.main_window_states |= Qt.WindowMinimized
        values = Qt.WindowMinimized
        super().__call__(parser=parser, namespace=namespace, values=values, option_string=option_string)


class MainWindowAlwaysOnTopAction(FlagAction):

    def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace, values: str | Sequence[Any] | None, option_string: str | None = ...) -> None:
        namespace.main_window_flags |= Qt.WindowStaysOnTopHint
        values = Qt.WindowStaysOnTopHint
        super().__call__(parser=parser, namespace=namespace, values=values, option_string=option_string)


class ClearAppSettingsAction(FlagAction):

    def __call__(self, parser: argparse.ArgumentParser, namespace: argparse.Namespace, values: str | Sequence[Any] | None, option_string: str | None = ...) -> None:
        parser.application.clear_settings()
        super().__call__(parser=parser, namespace=namespace, values=values, option_string=option_string)
        parser.exit(message="Cleared Application settings.")


class VersionFlagAction(FlagAction):

    def __init__(self,
                 option_strings,
                 version=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 helpus="show program's version number and exit"):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            helpus=helpus)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        version = self.version
        if version is None:
            version = parser.version
        formatter = parser._get_formatter()
        formatter.add_text(version)
        parser._print_message(formatter.format_help(), sys.stdout)
        parser.exit()


class AppArgParser(argparse.ArgumentParser):

    def __init__(self, application: "GidQtApplication", result_item: AppArgParserResult = None, console: RichConsole = None):
        self.application = application
        self.result_item = result_item or AppArgParserResult()
        self.console = console or RichConsole(soft_wrap=True)
        self.raw_argvs = list(self.application.arguments())
        self.result_item.exec_file = Path(self.raw_argvs.pop(0))
        super().__init__(prog=self.application.pretty_name)
        self.unparsable_args: list[str] = None
        self.setup()

    def setup(self):
        self.application.argument_parse_result = self.result_item
        self.version = str(self.application.version)
        self.setup_std_args()

    @property
    def actions(self):
        return self._actions

    def setup_std_args(self):

        self.add_argument("-v", "--version", action=VersionFlagAction, version=self.version)
        self.add_argument("-max", "--maximized", action=MainWindowMaximizedAction)
        self.add_argument("-min", "--minimized", action=MainWindowMinimizedAction)
        self.add_argument("-t", "--always-on-top", action=MainWindowAlwaysOnTopAction)
        self.add_argument("-c", "--clear-settings", action=ClearAppSettingsAction)

    def parse_app_args(self):
        _, self.unparsable_args = super().parse_known_args(self.raw_argvs, self.result_item)

    def _print_message(self, message: str, file: IO[str] | None = None) -> None:
        self.console.file = file

        self.console.rule()
        self.console.print(message)
        self.console.rule()


class WindowHolder(QObject):
    windows: dict[str, QWidget] = {}

    def __init__(self, parent=None) -> None:
        super().__init__(parent)

    def _determine_name(self, window: QWidget):
        if hasattr(window, "name"):
            return window.name

        object_name = window.objectName()
        if object_name:
            return object_name

        return window.__class__.__name__

    def _check_stored_windows_closed(self):
        for window in list(self.windows.values()):
            if window.isVisible() is False:
                self.remove_window(window)

    def add_window(self, window: QWidget):
        self._check_stored_windows_closed()
        name = self._determine_name(window)

        self.windows[name] = window
        try:
            window.close_signal.connect(self.remove_window)
        except AttributeError:
            pass

    @Slot(QWidget)
    def remove_window(self, window: QWidget):
        name = self._determine_name(window)
        try:
            del self.windows[name]
        except KeyError:
            pass

    def window_objects(self) -> list[QWidget]:
        return list(self.windows.values())

    def __getitem__(self, name: str) -> QWidget:
        self._check_stored_windows_closed()
        return self.windows[name]

    def __setitem__(self, name: str, window: QWidget) -> None:
        self._check_stored_windows_closed()
        self.add_window(window=window, name=name)

    def __delitem__(self, name: str) -> None:
        self._check_stored_windows_closed()
        del self.windows[name]

    def get(self, name: str, default=None) -> Optional[QWidget]:
        self._check_stored_windows_closed()
        return self.windows.get(name, default)

    def __len__(self) -> int:
        self._check_stored_windows_closed()
        return len(self.windows)

    def values(self):
        self._check_stored_windows_closed()
        return self.windows.values()

    def keys(self):
        self._check_stored_windows_closed()
        return self.windows.keys()

    def items(self):
        self._check_stored_windows_closed()
        return self.windows.items()

    def pop(self, name: str, default=None) -> Optional[QWidget]:
        self._check_stored_windows_closed()
        return self.windows.pop(name, default)

    def clear(self) -> None:
        self.windows.clear()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(parent={self.parent()!r})"


class GidQtApplication(QApplication):
    default_icon = QT_DEFAULT_APP_ICON_IMAGE

    def __init__(self,
                 argvs: list[str],
                 meta_info: "MetaInfo"):
        self.is_setup: bool = False
        super().__init__(argvs)
        self.setObjectName("Application")

        self.meta_info = meta_info

        self.main_window: GidBaseMainWindow = None
        self.sys_tray: GidBaseSysTray = None
        self.icon = None
        self.splash_screens: dict[str, QSplashScreen] = {"startup": None, "shutdown": None}
        self.argument_parse_result: AppArgParserResult = None
        self.extra_windows = WindowHolder()
        self._gui_thread_pool: ThreadPoolExecutor = None

    @property
    def is_dev(self) -> bool:
        return self.meta_info.is_dev

    @property
    def name(self) -> str:
        return self.meta_info.app_name

    @property
    def pretty_name(self) -> str:
        if self.meta_info.pretty_app_name:
            return self.meta_info.pretty_app_name
        return self.name

    @property
    def organization_name(self) -> str:
        return self.meta_info.app_author

    @property
    def version(self) -> VersionItem:
        return self.meta_info.version

    @property
    def url(self) -> URL:
        return self.meta_info.url

    @property
    def settings(self) -> QSettings:
        if self.is_setup is False:
            raise ApplicationNotSetupError(f"Unable to use 'settings' before {self!r} has been setup.")
        return QSettings()

    @property
    def gui_thread_pool(self) -> ThreadPoolExecutor:
        if self._gui_thread_pool is None:
            self._gui_thread_pool = ThreadPoolExecutor(thread_name_prefix="gui_thread")

        return self._gui_thread_pool

    @classmethod
    def set_pre_flags(cls, pre_flags: dict[Qt.ApplicationAttribute:bool]):
        if cls.instance() is not None:
            ApplicationExistsError(existing_application=cls.instance(), msg="Pre Flags can only be set, before an application is instantiated")

        if pre_flags is None:
            return

        for flag, value in pre_flags.items():
            cls.setAttribute(flag, value)

    def set_icon(self, icon=Union["PixmapResourceItem", QPixmap, QImage, str, QIcon, Path]):
        self.icon = self._icon_conversion(icon)
        self.setWindowIcon(self.icon)

    def setup(self) -> "GidQtApplication":
        if self.is_setup is False:
            self.setup_meta_data()
            self.additional_setup()
            self.is_setup = True
        return self

    def setup_meta_data(self):
        self.setApplicationName(self.name)
        self.setApplicationDisplayName(self.pretty_name)
        self.setOrganizationName(self.organization_name)
        if self.url:
            self.setOrganizationDomain(str(self.url))

        version = str(self.version) if self.version else "-"

        self.setApplicationVersion(version)

    def additional_setup(self) -> None:
        ...

    def clear_settings(self):
        self.settings.clear()

    def _icon_conversion(self, icon: Union["PixmapResourceItem", QPixmap, QImage, str, QIcon, Path] = None) -> Optional[QIcon]:
        if icon is None:
            icon = self.default_icon

        if isinstance(icon, QIcon):
            return icon

        if isinstance(icon, (QPixmap, QImage, str, Path)):
            if isinstance(icon, Path):
                icon = str(icon)
            return QIcon(icon)

        return icon.get_as_icon()

    def show_about_qt(self) -> None:
        self.aboutQt()

    def _get_about_text(self) -> str:
        template = Environment(loader=BaseLoader).from_string(ABOUT_TEMPLATE_FILE.read_text(encoding='utf-8', errors='ignore'))

        text_parts = {"Name": self.applicationDisplayName(),
                      "Author": self.organizationName(),
                      "Version": self.applicationVersion(),
                      "Dev Mode": "Yes" if self.is_dev is True else "No",
                      "Operating System": self.meta_info.os,
                      "Python Version": self.meta_info.python_version,
                      "License": self.meta_info.app_license,
                      "Summary": self.meta_info.summary,
                      "Description": self.meta_info.description}
        if self.organizationDomain():
            text_parts["link"] = f'<a href="{self.organizationDomain()}">{self.organizationDomain()}</a>'
        return template.render(style=ABOUT_STYLESHEET_FILE.read_text(encoding='utf-8', errors='ignore'), data=text_parts)

    def show_about(self) -> None:
        title = f"About {self.applicationDisplayName()}"
        text = self._get_about_text()
        QMessageBox.about(self.main_window, title, text)

    def start(self):
        self.setup()

        if self.splash_screens["startup"] and self.main_window:
            self.splash_screens["startup"].show()
            self.splash_screens["startup"].finish(self.main_window)

        if self.sys_tray:
            self.sys_tray.show()

        if self.main_window:
            self.main_window.show()
        return self.exec()

    def on_quit(self, event: QEvent):
        self.extra_windows.clear()
        self.closeAllWindows()
        if self._gui_thread_pool is not None:
            self._gui_thread_pool.shutdown(wait=False, cancel_futures=True)

    def event(self, event: PySide6.QtCore.QEvent) -> bool:
        if event.type() is QEvent.Quit:
            self.on_quit(event)

        return super().event(event)

    def __repr__(self) -> str:
        return make_repr(self, exclude_none=False, attr_names=["name", "organization_name", "version", "url", "arguments", "main_window", "sys_tray"])


T = TypeVar("T")


class ClassHolder:

    def __init__(self, klass: T, **kwargs) -> None:
        self.klass = klass
        self.kwargs = kwargs
        self.instance: T = None

    def create(self):
        self.instance = self.klass(**self.kwargs)
        return self.instance


class ApplicationBuilder:

    def __init__(self) -> None:
        self.application_pre_flags: dict[Qt.ApplicationAttribute:bool] = {Qt.AA_EnableHighDpiScaling: True, Qt.AA_UseHighDpiPixmaps: True}
        self.application_class: ClassHolder = ClassHolder(GidQtApplication)
        self.main_window_class: ClassHolder = ClassHolder(GidBaseMainWindow)
        self.argument_parser_class: ClassHolder = ClassHolder(AppArgParser)
        self.sys_tray_class: ClassHolder = None

        self.app_icon = None
        self.meta_info: "MetaInfo" = self._get_default_meta_info()
        self.style_sheet_file: Path = None

    def set_app_icon(self, icon: Union["PixmapResourceItem", QPixmap, QImage, str, QIcon]):
        self.app_icon = icon

    def set_application_class(self, application_class: type[GidQtApplication], **kwargs):
        self.application_class = ClassHolder(application_class, **kwargs)

    def set_main_window_class(self, main_window_class: type[QMainWindow], **kwargs):
        self.main_window_class = ClassHolder(main_window_class, **kwargs)

    def set_sys_tray_class(self, sys_tray_class: Union[type[QSystemTrayIcon], Literal[MiscEnum.DEFAULT], None] = MiscEnum.DEFAULT, **kwargs):
        if sys_tray_class is None:
            self.sys_tray_class = None
        elif sys_tray_class is MiscEnum.DEFAULT:
            self.sys_tray_class = ClassHolder(GidBaseSysTray, **kwargs)

        else:
            self.sys_tray_class = ClassHolder(sys_tray_class, **kwargs)

    def set_meta_info(self, meta_info: type[MetaInfo]):
        self.meta_info = meta_info

    def set_application_pre_flags(self, pre_flags: dict[Qt.ApplicationAttribute:bool]):
        self.application_pre_flags = pre_flags

    def set_style_sheet_file(self, file_path: Union[str, os.PathLike]):
        self.style_sheet_file = Path(file_path)

    def _get_default_meta_info(self):
        try:
            return get_meta_info()
        except (MetaItemNotFoundError, NotSetupError):
            return MetaInfo(app_name=Path(sys.argv[0]).stem.title())

    def _build_application(self) -> QApplication:
        self.application_class.klass.set_pre_flags(self.application_pre_flags)
        if "meta_info" not in self.application_class.kwargs:
            self.application_class.kwargs["meta_info"] = self.meta_info
        if "argvs" not in self.application_class.kwargs:
            self.application_class.kwargs["argvs"] = sys.argv
        application = self.application_class.create()
        application.set_icon(self.app_icon)
        return application

    def _build_argument_parser(self) -> AppArgParser:
        self.argument_parser_class.kwargs["application"] = self.application_class.instance
        parser = self.argument_parser_class.create()
        parser.parse_app_args()

        return parser

    def _build_main_window(self) -> QMainWindow:
        main_window: QMainWindow = self.main_window_class.create()
        main_window.setWindowFlags(main_window.windowFlags() | self.argument_parser_class.instance.result_item.main_window_flags)
        main_window.setWindowState(main_window.windowState() | self.argument_parser_class.instance.result_item.main_window_states)
        return main_window

    def _build_sys_tray(self) -> QSystemTrayIcon:
        if "icon" not in self.sys_tray_class.kwargs:
            self.sys_tray_class.kwargs["icon"] = self.application_class.instance.icon
        if "title" not in self.sys_tray_class.kwargs:
            self.sys_tray_class.kwargs["title"] = self.application_class.instance.pretty_name
        if "tooltip" not in self.sys_tray_class.kwargs:
            self.sys_tray_class.kwargs["tooltip"] = self.application_class.instance.pretty_name
        return self.sys_tray_class.create().setup()

    def build(self) -> GidQtApplication:
        application = self._build_application()
        parser = self._build_argument_parser()

        if self.sys_tray_class:
            sys_tray = self._build_sys_tray()

            application.sys_tray = sys_tray
        if self.main_window_class:

            main_window = self._build_main_window()
            application.main_window = main_window
        if self.style_sheet_file:
            application.setStyleSheet(self.style_sheet_file.read_text(encoding='utf-8', errors='ignore'))
        return application


# region [Main_Exec]
if __name__ == '__main__':

    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtWidgets import QWidget, QStatusBar, QMainWindow, QApplication

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import ApplicationNotExistingError
from gidapptools.gidapptools_qt.basics.menu_bar import BaseMenuBar
from gidapptools.gidapptools_qt.basics.status_bar import GidBaseStatusBar

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.basics.application import GidQtApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class GidBaseMainWindow(QMainWindow):

    def __init__(self, parent: Optional[PySide6.QtWidgets.QWidget] = None, flags: PySide6.QtCore.Qt.WindowFlags = None, menu_bar: BaseMenuBar = None, status_bar: GidBaseStatusBar = None) -> None:
        super().__init__(*(i for i in [parent, flags] if i))
        self.setObjectName("MainWindow")
        self.menu_bar = menu_bar or BaseMenuBar(auto_connect_standard_actions=True)
        self.status_bar = status_bar or GidBaseStatusBar()

    @property
    def app(self) -> "GidQtApplication":
        app = QApplication.instance()
        if app is None:
            raise ApplicationNotExistingError()
        return app

    @property
    def central_widget(self) -> QWidget:
        return super().centralWidget()

    @central_widget.setter
    def central_widget(self, widget: QWidget):
        self.setCentralWidget(widget)

    @property
    def menu_bar(self) -> BaseMenuBar:
        return super().menuBar()

    @menu_bar.setter
    def menu_bar(self, menu_bar: BaseMenuBar):
        menu_bar.setParent(self)
        try:
            menu_bar.setup()
        except AttributeError:
            pass
        self.setMenuBar(menu_bar)

    @property
    def status_bar(self) -> QStatusBar:
        return super().statusBar()

    @status_bar.setter
    def status_bar(self, status_bar: QStatusBar):
        status_bar.setParent(self)
        try:
            status_bar.setup()
        except AttributeError:
            pass
        self.setStatusBar(status_bar)
# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
import enum
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
import types
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final, Callable,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from gidapptools.gidcolor.color import Color

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

FORMATTER_TYPE: TypeAlias = Callable[["Color", bool], str]


FORMATTER_TYPE_T = TypeVar("FORMATTER_TYPE_T", bound=FORMATTER_TYPE)


@enum.verify(enum.UNIQUE)
class ColorFormatSystem(enum.Enum):
    RGB = "rgb"
    HSL = "hsl"
    HSV = "hsv"

    HLS = "hls"

    @classmethod
    @property
    def target_string(self) -> str:
        return "system"


@enum.verify(enum.UNIQUE)
class ColorFormatContext(enum.Enum):
    CSS = "css"

    @classmethod
    @property
    def target_string(self) -> str:
        return "context"


@enum.verify(enum.UNIQUE)
class ColorFormatTypus(enum.Enum):
    FLOAT = "float"
    INT = "int"

    PRETTY = "pretty"

    @classmethod
    @property
    def target_string(self) -> str:
        return "typus"


class ColorStringFormatter:

    alpha_part_regex = re.compile(r"[+-](a|alpha)", re.IGNORECASE)

    def __init__(self) -> None:
        pass

    def parse_format_string(self, format_string: str) -> tuple["ColorFormat", bool]:
        color_format = ColorFormat()
        with_alpha = True
        for format_enum in (ColorFormatSystem, ColorFormatContext, ColorFormatTypus):
            if format_string.startswith(("+a", "-a")):
                with_alpha = format_string[0] == "+"
                format_string = format_string[2:]
            for item in format_enum._member_map_.values():
                identifier = item.value

                if format_string.startswith(identifier):
                    print(identifier)
                    setattr(color_format, format_enum.target_string, item)
                    format_string = format_string.removeprefix(identifier)
                    break

        if format_string != "":
            raise ValueError(f"Invalid Format-String {format_string!r}")
        return color_format, with_alpha

    def format_color(self,
                     color: "Color",
                     color_format: "ColorFormat" = None,
                     with_alpha: bool = True) -> str:
        if color_format is None:
            color_format = ColorFormat()

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}'


class ColorFormat:

    def __init__(self,
                 system: ColorFormatSystem = None,
                 context: ColorFormatContext = None,
                 typus: ColorFormatTypus = None) -> None:
        self.system = system
        self.context = context
        self.typus = typus

    def __or__(self, other: object) -> Self:
        if isinstance(other, ColorFormatSystem):
            return self.__class__(system=other, context=self.context, typus=self.typus)
        if isinstance(other, ColorFormatContext):
            return self.__class__(system=self.system, context=other, typus=self.typus)

        if isinstance(other, ColorFormatTypus):
            return self.__class__(system=self.system, context=self.context, typus=other)

        else:
            return NotImplemented

    def __contains__(self, other: object) -> bool:
        if isinstance(other, ColorFormatSystem):
            return self.system is other
        if isinstance(other, ColorFormatContext):
            return self.context is other

        if isinstance(other, ColorFormatTypus):
            return self.typus is other

        else:
            return NotImplemented

    def __str__(self) -> str:
        parts = []
        for attr_name in ("system", "context", "typus"):
            value = getattr(self, attr_name)
            if value is None:
                parts.append(str(None))
            else:
                parts.append(value.name)

        return "|".join(parts)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(system={self.system!r}, context={self.context!r}, typus={self.typus!r})'

# region [Main_Exec]


if __name__ == '__main__':
    x = ColorStringFormatter()

    y = "rgb+acssfloat"

    u = x.parse_format_string(y)

    print(u)

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import inspect
import logging
import threading
from enum import Enum
from typing import Any, Union, Callable
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_logger.enums import LoggingLevel

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
logging.basicConfig()
log = logging.getLogger(__name__)

log.setLevel(LoggingLevel.DEBUG)
# endregion [Constants]


class GidLogRecordFactory:
    activation_lock = threading.RLock()
    __slots__ = ("special_records_registry",
                 "original_factory")

    def __init__(self) -> None:
        self.special_records_registry: dict[str, logging.LogRecord] = {}
        self.original_factory: Callable = None

    def is_active(self) -> bool:
        return logging.getLogRecordFactory() is self

    def __call__(self,
                 name: str,
                 level: Union[int, str, Enum],
                 pathname: Union[os.PathLike, str, Path],
                 lineno: int,
                 msg: str,
                 args: tuple[Any],
                 exc_info: Union[tuple, bool, BaseException] = None,
                 func: str = None,
                 sinfo: str = None,
                 **kwargs):
        special_record = kwargs.pop("record_typus", None)

        if inspect.isclass(special_record):
            record_class = special_record
        else:
            record_class = self.special_records_registry.get(special_record, self.original_factory)

        return record_class(name=name,
                            level=level,
                            pathname=pathname,
                            lineno=lineno,
                            msg=msg,
                            args=args,
                            exc_info=exc_info,
                            func=func,
                            sinfo=sinfo,
                            **kwargs)

    def activate(self, raise_on_already_active: bool = False) -> bool:
        with self.activation_lock:

            if self.is_active() is True:
                if raise_on_already_active is True:
                    # TODO: Custom Error!
                    raise RuntimeError(f"{self!r} is already active.")
                return False

            self.original_factory = logging.getLogRecordFactory()
            logging.setLogRecordFactory(self)
            return True

    def deactivate(self, raise_on_not_active: bool = False) -> bool:
        with self.activation_lock:

            if self.is_active() is False:
                if raise_on_not_active is True:
                    # TODO: Custom Error!
                    raise RuntimeError(f"Unable to deactivate {self!r}, because it is currently not active.")
                return False

            logging.setLogRecordFactory(self.original_factory)

            self.original_factory = None
            return True


gid_log_record_factory = GidLogRecordFactory()


class GidBaseLogRecord(logging.LogRecord):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.extras = {}


LOG_RECORD_TYPES = Union[logging.LogRecord, GidBaseLogRecord]

# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import sys
import platform
from typing import TYPE_CHECKING, Any, Union, Callable, Optional
from pathlib import Path
from datetime import datetime, timezone
from functools import cached_property, partial
import psutil
# * Third Party Imports --------------------------------------------------------------------------------->
import attr
from yarl import URL
from tzlocal import get_localzone

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_utility import VersionItem
from gidapptools.utility.enums import OperatingSystem
from gidapptools.utility.helper import utc_now, make_pretty, memory_in_use
from gidapptools.general_helper.general import is_frozen
from gidapptools.general_helper.date_time import DatetimeFmt
from gidapptools.general_helper.conversion import bytes2human
from gidapptools.general_helper.string_helper import StringCase, StringCaseConverter
from gidapptools.abstract_classes.abstract_meta_item import AbstractMetaItem

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.meta_data.meta_info.meta_info_factory import License

# endregion [Imports]

# region [TODO]

# - Make into a class

# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion [Constants]

def url_converter(in_url: Optional[str]) -> Optional[URL]:
    if in_url is None:
        return in_url
    return URL(in_url)


def many_url_converter(in_urls: Optional[dict[str, Union[str, URL]]]) -> Optional[dict[str, URL]]:
    if in_urls is None:
        return in_urls
    return {k: URL(v) for k, v in in_urls.items()}


def version_converter(in_version: Optional[Union[str, VersionItem]]) -> VersionItem:
    if in_version is None:
        return None
    if isinstance(in_version, VersionItem):
        return in_version
    return VersionItem.from_string(in_version)


@attr.s(auto_attribs=True, auto_detect=True, kw_only=True, frozen=True)
class MetaInfo(AbstractMetaItem):
    app_name: str = attr.ib(default="python_script")
    app_author: str = attr.ib(default=None)
    version: VersionItem = attr.ib(default=None, converter=version_converter)
    url: URL = attr.ib(converter=url_converter, default=None)
    other_urls: dict[str, URL] = attr.ib(converter=many_url_converter, default=dict())
    pid: int = attr.ib(factory=os.getpid)
    process: psutil.Process = attr.ib(factory=partial(psutil.Process, os.getpid()))
    os: OperatingSystem = attr.ib(factory=OperatingSystem.determine_operating_system)
    os_release: str = attr.ib(factory=platform.release)
    python_version: VersionItem = attr.ib(factory=platform.python_version, converter=version_converter)
    is_dev: bool = attr.ib(default=None, converter=attr.converters.default_if_none(False))
    is_gui: bool = attr.ib(default=None, converter=attr.converters.default_if_none(False))
    local_tz: timezone = attr.ib(default=get_localzone())
    summary: str = attr.ib(default=None)
    app_license: "License" = attr.ib(default=None)
    description: str = attr.ib(default=None)

    @property
    def started_at(self) -> datetime:
        time_stamp = self.process.create_time()
        return datetime.fromtimestamp(time_stamp, tz=timezone.utc)

    @property
    def is_frozen_app(self) -> bool:
        return is_frozen()

    @property
    def frozen_folder_path(self) -> Optional[Path]:
        if self.is_frozen_app is True:
            return Path(sys._MEIPASS)

    @property
    def cli_name(self) -> str:
        return self.app_name.replace(" ", "-").replace("_", '-')

    @classmethod
    @property
    def __default_configuration__(cls) -> dict[str, Any]:
        default_configuration = {}
        return default_configuration

    @property
    def pretty_is_dev(self) -> str:
        return "Yes" if self.is_dev else "No"

    @property
    def pretty_app_name(self) -> str:
        if self.app_name:
            return StringCaseConverter.convert_to(self.app_name, StringCase.TITLE)

    @property
    def pretty_app_author(self) -> str:
        if self.app_author:
            return StringCaseConverter.convert_to(self.app_author, StringCase.TITLE)

    @property
    def pretty_started_at(self) -> str:
        return self.started_at.strftime(DatetimeFmt.STANDARD)

    def as_dict(self, pretty: bool = False) -> dict[str, Any]:
        if pretty is True:
            return make_pretty(self)
        return attr.asdict(self)

    def to_storager(self, storager: Callable = None) -> None:
        if storager is None:
            return
        storager(self)

    def clean_up(self, **kwargs) -> None:
        pass


def ManualMetaInfoItem(app_name: str = None,
                       app_author: str = None,
                       version: Union[str, VersionItem] = None,
                       url: Union[str, URL] = None,
                       other_urls: dict[str, Union[str, URL]] = None,
                       is_dev: bool = None,
                       is_gui: bool = None,
                       summary: str = None,
                       description: str = None) -> MetaInfo:

    is_dev = is_dev if is_dev is not None else os.getenv("IS_DEV", "0") != "0"
    kwargs = dict(app_name=app_name, app_author=app_author, version=version, url=url, other_urls=other_urls, is_dev=is_dev, is_gui=is_gui, summary=summary, description=description)
    for k in list(kwargs):
        if kwargs[k] is None:
            kwargs.pop(k)
    return MetaInfo(**kwargs)

    # region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import string
import subprocess
from glob import iglob
from ctypes import windll
from typing import TYPE_CHECKING, Union, Optional, Generator
from pathlib import Path
from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor

# * Third Party Imports --------------------------------------------------------------------------------->
from psutil import disk_partitions

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.enums import OperatingSystem

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def get_all_drives(also_non_physical: bool = False) -> tuple[Path]:
    return tuple(Path(drive.mountpoint) for drive in disk_partitions(all=also_non_physical))


def get_all_drives_non_psutil(*args):
    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            drives.append(Path(f"{letter}:\\ ").resolve())
        bitmask >>= 1

    return tuple(drives)


def find_file(file_name: str, return_first: bool = True, case_sensitive: bool = False) -> Optional[Union[Path, tuple[Path]]]:
    if case_sensitive is False:
        file_name = file_name.casefold()
    _out = []
    for drive in get_all_drives():
        for dirname, folderlist, filelist in os.walk(drive):
            for file in filelist:
                if file.casefold() == file_name:
                    file_path = Path(dirname, file)
                    if return_first is True:
                        return file_path
                    else:
                        _out.append(file_path)
    if _out != []:
        return tuple(_out)
    return None


def find_file_alternative(file_name: str, return_first: bool = True) -> Optional[Union[Path, tuple[Path]]]:
    def _helper(args) -> Generator[Path, None, None]:
        _drive = args[0]
        _file_name = args[1]
        for _found_file in iglob(f"{_drive.as_posix().rstrip('/')}/**/{_file_name}", recursive=True):
            yield Path(_found_file)
    _out = []
    with ThreadPoolExecutor(3) as pool:
        for gen in pool.map(_helper, ((drive, file_name) for drive in get_all_drives())):
            for file in gen:
                if return_first is True:
                    return file
                else:
                    _out.append(file)
    if _out != []:
        return tuple(_out)
    return None


@contextmanager
def change_cwd(target_cwd: "PATH_TYPE"):
    old_cwd = Path.cwd()
    new_cwd = Path(target_cwd)
    if new_cwd.is_dir() is False:
        raise FileNotFoundError(f"The target_cwd({new_cwd.as_posix()!r}) either does not exist or is a file and not a directory.")
    os.chdir(new_cwd)
    yield
    os.chdir(old_cwd)


def open_folder_in_explorer(in_folder: Union[str, os.PathLike]) -> None:
    in_folder = Path(in_folder)
    if not in_folder.exists():
        raise FileNotFoundError(f"No such file or directory: {in_folder.as_posix()!r}.")
    if not in_folder.is_dir():
        raise TypeError(f"Path {in_folder.as_posix()!r} needs to be a folder.")

    operating_system = OperatingSystem.determine_operating_system()

    match operating_system:

        case OperatingSystem.WINDOWS:
            subprocess.run(["explorer", in_folder], check=False, start_new_session=True)

        case OperatingSystem.LINUX:
            subprocess.run(['xdg-open', in_folder], check=False)

        case OperatingSystem.MAC_OS:
            subprocess.run(['open', in_folder], check=False)

        case _:
            raise RuntimeError(f"Not able to open folder {in_folder.as_posix()!r}, because no known procedure for Operating System {operating_system!s}.")


ILLEGAL_FILE_NAME_CHARS: set[str] = set("\"|%:/,.\\[]<>*?")

EXTENDED_ILLEGAL_FILE_NAME_CHARS: set[str] = ILLEGAL_FILE_NAME_CHARS.union("'&§\{\};#=")


def ensure_valid_file_stem(file_stem: str,
                           strict: bool = False,
                           replace_all_spaces: bool = False) -> str:
    illegal_chars = ILLEGAL_FILE_NAME_CHARS
    file_stem = file_stem.rstrip()

    if strict is True:
        illegal_chars = EXTENDED_ILLEGAL_FILE_NAME_CHARS
        file_stem = file_stem.lstrip()

    if replace_all_spaces is True:
        file_stem = file_stem.replace(" ", "_")

    return "".join(c for c in file_stem if c not in illegal_chars)


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
import types
import random
import shutil
import subprocess
from io import StringIO
from time import sleep
from string import ascii_letters
from typing import Any
from pathlib import Path
from tempfile import TemporaryDirectory

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import MissingOptionalDependencyError
from gidapptools.gid_logger.logger import get_logger
with MissingOptionalDependencyError.try_import("gidapptools"):
    from rich import inspect as rinspect
    from rich.tree import Tree
    from rich.panel import Panel
    from rich.console import Console as RichConsole
    from rich.terminal_theme import TerminalTheme

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
log = get_logger(__name__)
# endregion [Constants]

MY_TERMINAL_THEME = TerminalTheme(
    (40, 40, 40),
    (102, 255, 50),
    [
        (0, 0, 0),
        (15, 140, 220),
        (86, 216, 86),
        (255, 215, 0),
        (98, 96, 180),
        (18, 255, 156),
        (25, 128, 128),
        (192, 192, 192),
    ],
    [
        (128, 128, 128),
        (150, 25, 25),
        (200, 200, 100),
        (150, 150, 25),
        (25, 25, 150),
        (0, 176, 255),
        (213, 121, 255),
        (150, 150, 150),
    ],
)


def dict_to_rich_tree(label: str, in_dict: dict) -> Tree:
    base_tree = Tree(label=label)

    def _handle_sub_dict(in_sub_dict: dict, attach_node: Tree):
        for k, v in in_sub_dict.items():
            key_node = attach_node.add(k)
            if isinstance(v, dict):
                _handle_sub_dict(v, key_node)
            elif isinstance(v, list):
                key_node.add(Panel(',\n'.join(f"{i}" for i in v)))
            else:
                key_node.add(f"{v}")

    _handle_sub_dict(in_dict, base_tree)
    return base_tree


def inspect_object_with_html(obj: object,
                             show_all: bool = False,
                             show_methods: bool = False,
                             show_dunder: bool = False,
                             show_private: bool = False,
                             show_docs: bool = True,
                             show_help: bool = False):

    def _make_title(_obj: Any) -> str:

        title_str = (
            str(_obj)
            if (isinstance(_obj, type) or callable(_obj) or isinstance(_obj, types.ModuleType))
            else str(type(_obj))
        )
        if hasattr(obj, "name") and obj.name != str(_obj):
            title_str += f' -| {_obj.name!r} |-'

        return title_str

    def sanitize_name(name: str) -> str:

        return re.sub(r"\.\s\-\?\!\,\(\)\[\]\<\>\|\:\;\'\"\&\%\$\§\\", '_', name)

    def make_file_name(_obj) -> str:
        # if hasattr(_obj, 'name'):
        #     text = _obj.name

        # elif hasattr(_obj, '__name__'):
        #     text = _obj.__name__
        # else:
        text = ''.join(random.choices(ascii_letters, k=random.randint(5, 10)))

        return sanitize_name(text) + '.html'

    with StringIO() as throw_away_file:
        console = RichConsole(soft_wrap=True, record=True, file=throw_away_file)
        title = None
        try:
            title = _make_title(obj)
        except Exception as e:
            log.error(e, exc_info=True)
            title = None
        rinspect(obj=obj,
                 title=title,
                 help=show_help,
                 methods=show_methods,
                 docs=show_docs,
                 private=show_private,
                 dunder=show_dunder,
                 all=show_all,
                 console=console)
        with TemporaryDirectory() as temp_directory:
            out_file = Path(temp_directory).joinpath(make_file_name(obj))

            console.save_html(out_file, theme=MY_TERMINAL_THEME)
            firefox = shutil.which("firefox.exe")
            cmd = f'"{firefox}" -new-window "{str(out_file)}"'

            _cmd = subprocess.run(cmd, check=True, text=True)
            sleep(0.5)


# region [Main_Exec]

if __name__ == '__main__':

    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as pp

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BaseElements:
    comma = pp.Literal(",").suppress()
    colon = pp.Literal(":").suppress()
    semi_colon = pp.Literal(";").suppress()
    period = pp.Literal(".").suppress()
    pipe = pp.Literal("|").suppress()
    at = pp.Literal("@").suppress()
    hyhphen = pp.Literal("-").suppress()

    octothorp = pp.Literal("#").suppress()
    tilde = pp.Literal("~").suppress()

    plus = pp.Literal("+").suppress()
    minus = pp.Literal("-").suppress()
    asterisk = pp.Literal("*").suppress()
    equals = pp.Literal("=").suppress()

    forward_slash = pp.Literal("/").suppress()
    back_slash = pp.Literal("/").suppress()

    single_quote = pp.Literal("'").suppress()
    double_quote = pp.Literal('"').suppress()
    any_quote = single_quote | double_quote

    parentheses_open = pp.Literal("(").suppress()
    parentheses_close = pp.Literal(")").suppress()

    brackets_open = pp.Literal("[").suppress()
    brackets_close = pp.Literal("]").suppress()

    braces_open = pp.Literal("{").suppress()
    braces_close = pp.Literal("}").suppress()


class Ligatures:
    arrow_right = pp.Literal("->").suppress()
    arrow_left = pp.Literal("<-").suppress()

    big_arrow_right = pp.Literal("-->").suppress()
    big_arrow_left = pp.Literal("<--").suppress()


COMMA = BaseElements.comma
COLON = BaseElements.colon
SEMI_COLON = BaseElements.semi_colon
PERIOD = BaseElements.period
PIPE = BaseElements.pipe
AT = BaseElements.at
HYHPHEN = BaseElements.hyhphen
OCTOTHORP = BaseElements.octothorp
TILDE = BaseElements.tilde
PLUS = BaseElements.plus
MINUS = BaseElements.minus
ASTERISK = BaseElements.asterisk
EQUALS = BaseElements.equals
FORWARD_SLASH = BaseElements.forward_slash
BACK_SLASH = BaseElements.back_slash
SINGLE_QUOTE = BaseElements.single_quote
DOUBLE_QUOTE = BaseElements.double_quote
ANY_QUOTE = BaseElements.any_quote
PARENTHESES_OPEN = BaseElements.parentheses_open
PARENTHESES_CLOSE = BaseElements.parentheses_close
BRACKETS_OPEN = BaseElements.brackets_open
BRACKETS_CLOSE = BaseElements.brackets_close
BRACES_OPEN = BaseElements.braces_open
BRACES_CLOSE = BaseElements.braces_close


ARROW_RIGHT = Ligatures.arrow_right
ARROW_LEFT = Ligatures.arrow_left
BIG_ARROW_RIGHT = Ligatures.big_arrow_right
BIG_ARROW_LEFT = Ligatures.big_arrow_left

# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from abc import ABC, abstractmethod
from typing import Any, Callable
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.string_helper import StringCaseConverter

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class AbstractMetaItem(ABC):

    @classmethod
    @property
    def __default_configuration__(cls) -> dict[str, Any]:
        return {}

    @classmethod
    @property
    def name(cls) -> str:
        return StringCaseConverter.convert_to(cls.__name__, StringCaseConverter.SNAKE)

    @abstractmethod
    def as_dict(self, pretty: bool = False) -> dict[str, Any]:
        ...

    @abstractmethod
    def to_storager(self, storager: Callable = None) -> None:
        ...

    @abstractmethod
    def clean_up(self, **kwargs) -> None:
        ...


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]
